! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!              Helper for MPI_Type_create_struct               !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Helper for MPI_Type_create_struct
!!
!!>
!!  call fd4_mpitype_add(mdt, index, blen, displ, ...)    ! add a memory location to the type
!!  call fd4_mpitype_add(mdt, index, blen, displ, ...)    ! add a memory location to the type
!!  call fd4_mpitype_add(mdt, index, blen, displ, ...)    ! add a memory location to the type
!!  ...
!!  call fd4_mpitype_commit(mdt, err)           ! call mpi_type_create_struct and ..._commit
!!  ...
!!  call MPI_Send( ... , mdt%datatype ... )
!!  ...
!!  call fd4_mpitype_free(mdt)
!!<
!!
!! Note: $index is an 8 byte integer!
!!
!! Note: $sortmdt is a global variable which will be allocated, but not deallocated. This is intented behavior.
!!
!! @author Matthias Lieber
module fd4_mpitype_mod

  use fd4_globaldef_mod
  use mrgrnk_mod

  USE_MPI
  
  implicit none
  private

  INCLUDE_MPIFH
  
  public :: fd4_mpitype_add, fd4_mpitype_allocate, fd4_mpitype_free, fd4_mpitype_commit
  public :: fd4_mpitype

  type fd4_mpitype
    integer :: datatype = MPI_DATATYPE_NULL
    integer(i8k), pointer :: indexes(:) => null()
    integer, pointer :: blens(:) => null()
    integer, pointer :: types(:) => null()
    integer(MPI_ADDRESS_KIND), pointer :: displ(:) => null()
    integer(MPI_ADDRESS_KIND) :: mindispl = -1
    integer :: count = 0
    integer(i_k), pointer :: sort(:) => null() !! used in sortmdt only
  end type fd4_mpitype

  !! temp variable for sorting, used by all $fd4_mpitypes
  !! its data will be deallocated nowhere at program end
  type(fd4_mpitype), save, public :: sortmdt


contains

  
  !! Add a component to the $fd4_mpitype mdt.
  !!
  !! Use the $is_min parameter if you need to know the component with the lowest displacement
  !! This is the one you should use in mpi_send etc. when using the resulting mpi datatype.
  !! $fd4_mpitype_commit calculates the displacements relative to the minimal displacement
  !! to avoid negative displacements.
  !! 
  !! The index is used for sorting in $fd4_mpitype_commit to allow sender-receiver
  !! matching of the types.
  subroutine fd4_mpitype_add(mdt, index, blen, displ, mpitype, is_min)
    type(fd4_mpitype), intent(inout) :: mdt      !! the $fd4_mpitype
    integer(i8k), intent(in)             :: index    !! index used for sorting
    integer, intent(in)                  :: blen     !! block length for constructing datatype
    integer(MPI_ADDRESS_KIND),intent(in) :: displ    !! displacement for constructing datatype
    integer, intent(in)                  :: mpitype  !! type for constructing datatype
    logical, intent(out), optional       :: is_min   !! true if the displacement is the lowest of (currently) all
    
    integer(i_k), parameter :: allocate_step = 32
    
    ! ensure mdt's fields are allocated large enough
    if(.not. associated(mdt%indexes)) then
      call fd4_mpitype_allocate(mdt, allocate_step)
    else if(mdt%count >= ubound(mdt%indexes,1)) then
      call fd4_mpitype_allocate(mdt, mdt%count + allocate_step)
    end if
    
    mdt%count = mdt%count + 1
    mdt%indexes(mdt%count) = index
    mdt%blens(mdt%count)   = blen
    mdt%types(mdt%count)   = mpitype
    mdt%displ(mdt%count)   = displ
    if( mdt%mindispl > displ .or. mdt%count == 1 ) then
      mdt%mindispl = displ
      if(present(is_min)) is_min = .true.
    else if(present(is_min)) then
      is_min=.false.
    end if
      
  end subroutine fd4_mpitype_add


  !! Construct and commit the MPI dataype.
  !! The MPI datatype is stored in $mdt%datatype.
  !! 
  !! Notes about $startdispl:
  !!  - address of the buffer used for the MPI send/recv call
  !!  - address is subtracted from all component's displacements
  !!  - default is to use the min displacement of all components (use $is_min of $fd4_mpitype_add)
  !!  - use 0 for $startdispl when using $MPI_BOTTOM
  !!
  !! *TODO*:
  !!  - option to not sort the components?
  subroutine fd4_mpitype_commit(mdt, err, startdispl)
    type(fd4_mpitype),  intent(inout) :: mdt      !! the $fd4_mpitype
    integer(i_k), intent(out)         :: err      !! error status: 0...ok
    integer(MPI_ADDRESS_KIND), intent(in), optional :: startdispl !! start address of the type
    integer(i_k) :: i

    err = -1
    
    if(mdt%count<=0 .or. (.not. associated(mdt%indexes)) ) then
      FD4_L1_ERROR('empty fd4_mpitype')
    end if
    
    if(present(startdispl)) mdt%mindispl = startdispl
    
    ! ensure global variable sortmdt is large enough
    if(.not. associated(sortmdt%blens)) then
      call fd4_mpitype_allocate(sortmdt, mdt%count, .true.)
    else if(mdt%count >= ubound(sortmdt%blens,1)) then
      call fd4_mpitype_allocate(sortmdt, mdt%count, .true.)
    end if

    ! rank mdt%indexes
    call mrgrnk(mdt%indexes(1:mdt%count),sortmdt%sort(1:mdt%count))
    
    ! sortmdt%blens contains the ranks of mdt%indexes and is used for sorting blens, types, and displ
    do i=1,mdt%count
      sortmdt%blens(i) = mdt%blens(sortmdt%sort(i))
    end do
    do i=1,mdt%count
      sortmdt%types(i) = mdt%types(sortmdt%sort(i))
    end do
    do i=1,mdt%count
      sortmdt%displ(i) = mdt%displ(sortmdt%sort(i)) - mdt%mindispl
    end do

    ! construct the mpi datatype, using the sorted arrays blens, types, and displ of sortmdt
    call MPI_Type_create_struct(mdt%count, sortmdt%blens, sortmdt%displ, sortmdt%types, mdt%datatype, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Type_create_struct failed')
    call MPI_Type_commit(mdt%datatype, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Type_commit failed')
    
    err = 0

  end subroutine fd4_mpitype_commit




  !! Allocate an $fd4_mpitype.
  !!  - allocate if $mdt is not yet allocated
  !!  - deallocate all memory if $length = 0
  !!  - otherwise reallocate, copy (1:count) from old to new memory
  subroutine fd4_mpitype_allocate(mdt, length, is_sortmdt)
    type(fd4_mpitype),  intent(inout) :: mdt        !! the $fd4_mpitype
    integer(i_k), intent(in)          :: length     !! size to allocate (max number of components)
    logical, optional, intent(in)     :: is_sortmdt !! internal option only, do not use!
    type(fd4_mpitype) :: tmp
    
#ifdef FD4_GFORTRAN
    ! gfortran bug workaround
    tmp%indexes => null()
    tmp%blens   => null()
    tmp%types   => null()
    tmp%displ   => null()
    tmp%sort    => null()
#endif

    if(length > 0) then
      if(present(is_sortmdt)) then
        allocate(tmp%sort(length))
      else
        allocate(tmp%indexes(length))
      end if
      allocate(tmp%blens(length))
      allocate(tmp%types(length))
      tmp%types = MPI_DATATYPE_NULL
      allocate(tmp%displ(length))
      if(associated(mdt%indexes) .and. mdt%count>0) then
        if(.not. present(is_sortmdt)) &
          tmp%indexes(1:mdt%count) = mdt%indexes(1:mdt%count) 
        tmp%blens(1:mdt%count)   = mdt%blens(1:mdt%count) 
        tmp%types(1:mdt%count)   = mdt%types(1:mdt%count) 
        tmp%displ(1:mdt%count)   = mdt%displ(1:mdt%count)
      end if
    end if

    if(associated(mdt%blens)) then
      if(present(is_sortmdt)) then
        deallocate(mdt%sort)
      else
        deallocate(mdt%indexes)
      end if
      deallocate(mdt%blens)
      deallocate(mdt%types)
      deallocate(mdt%displ)
    end if
    
    if(present(is_sortmdt)) then
      mdt%sort    => tmp%sort
    else
      mdt%indexes => tmp%indexes
    end if
    mdt%blens   => tmp%blens
    mdt%types   => tmp%types
    mdt%displ   => tmp%displ
        
  end subroutine fd4_mpitype_allocate


  !! Free the mpi datatype, reset the component counter to zero and, if $free_memory is true, deallocate all memory.
  !! 
  !! $free_memory does *not* imply $free_types or vice versa!
  subroutine fd4_mpitype_free(mdt, free_memory, free_types)
    type(fd4_mpitype),  intent(inout) :: mdt           !! the $fd4_mpitype
    logical, intent(in)               :: free_memory   !! free the arrays used to construct the MPI datatype
    logical, intent(in)               :: free_types    !! free the basic types used to construct the MPI datatype
    
    integer :: err, i

    if(free_types) then
      do i=1,mdt%count
        if(mdt%types(i) /= MPI_DATATYPE_NULL) then
          call MPI_Type_free(mdt%types(i), err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
          mdt%types(i) = MPI_DATATYPE_NULL
        end if
      end do
    end if

    mdt%count = 0
    mdt%mindispl = -1
    if(mdt%datatype /= MPI_DATATYPE_NULL) then
      call MPI_Type_free(mdt%datatype, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
      mdt%datatype = MPI_DATATYPE_NULL
    end if

    if(free_memory) call fd4_mpitype_allocate(mdt,0)

  end subroutine fd4_mpitype_free
  


end module fd4_mpitype_mod
