! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!            Data related with a neighbor partition            !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Data related with communication to a neighbor partition.
!! Used in ghost communication, coupling, and block migration.
!!
!! @author Matthias Lieber
module fd4_neigh_mod

  use fd4_block_mod, only: fd4_r_k_link
  use fd4_globaldef_mod
  use fd4_mpitype_mod
  use kinds_mod
  
  implicit none
  private
  
  public fd4_neigh, fd4_neigh_deallocate, fd4_neigh_deallocate_ptr, fd4_neigh_deallocate_type_ptr
  public fd4_mpitype, fd4_r_k_link, fd4_i4k_link
  public FD4_NEIGH_SEND, FD4_NEIGH_RECV
  
  integer(i_k), parameter :: FD4_NEIGH_SEND = 1
  integer(i_k), parameter :: FD4_NEIGH_RECV = 2

  type fd4_i4k_link
    integer(i4k), pointer :: l => NULL()
  end type fd4_i4k_link
  
  type fd4_neigh
    ! stuff for constructing MPI datatypes
    type(fd4_mpitype) :: mpitype(2)  !! MPI data type for send and receive
    type(fd4_r_k_link) :: minptr(2)  !! pointer to type component with min address (real)
    type(fd4_i4k_link) :: minpti(2)  !! pointer to type component with min address (integer)
    logical :: changed = .false.     !! flag to notify that neigh has been changed
  end type fd4_neigh

contains

  !! Deallocate all data of the $fd4_neigh.
  subroutine fd4_neigh_deallocate(n)
    type(fd4_neigh), intent(inout) :: n          !! $fd4_neigh to deallocate
    call fd4_mpitype_free(n%mpitype( FD4_NEIGH_SEND), .true., .false. )
    call fd4_mpitype_free(n%mpitype( FD4_NEIGH_RECV), .true., .false. )
    NULLIFY( n%minptr(FD4_NEIGH_SEND)%l )
    NULLIFY( n%minptr(FD4_NEIGH_RECV)%l )
  end subroutine fd4_neigh_deallocate

  !! Deallocate all data of the $fd4_neigh, pointer version.
  !! Deallocates also the $fd4_neigh itself.
  subroutine fd4_neigh_deallocate_ptr(n)
    type(fd4_neigh), pointer :: n          !! pointer to $fd4_neigh to deallocate
    call fd4_mpitype_free(n%mpitype( FD4_NEIGH_SEND), .true., .false. )
    call fd4_mpitype_free(n%mpitype( FD4_NEIGH_RECV), .true., .false. )
    deallocate(n)
    NULLIFY(n)
  end subroutine fd4_neigh_deallocate_ptr
  
  !! Deallocate all data of the $fd4_neigh, pointer version.
  !! Deallocates also the $fd4_neigh itself and the basic types.
  subroutine fd4_neigh_deallocate_type_ptr(n)
    type(fd4_neigh), pointer :: n          !! pointer to $fd4_neigh to deallocate
    call fd4_mpitype_free(n%mpitype( FD4_NEIGH_SEND), .true., .true. )
    call fd4_mpitype_free(n%mpitype( FD4_NEIGH_RECV), .true., .true. )
    deallocate(n)
    NULLIFY(n)
  end subroutine fd4_neigh_deallocate_type_ptr

end module fd4_neigh_mod
