! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
! Distributed & Hierarchical Space-filling curve partitioning  !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"


!! Space-filling curve partitioning of an $fd4_domain.
!!
!! *EXPERIMENTAL DISTRIBUTED/HIERARCHICAL VERSION*
!!
!! Supports Hilbert and Morton SFC.
!! One-dimensional partitioning algorithms from $fd4_part_1d_mod are used.
!!
!! @see hilbert_mod
!! @see fd4_part_1d_mod
!! @see fd4_domain_mod
!! @see fd4_part_sfc_mod
!! @author Matthias Lieber
module fd4_part_sfcd_mod

  use fd4_globaldef_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_part_1d_mod
  use fd4_part_sfc_mod
  use fd4_mpi_mod
  use hilbert_mod
  use hilbert2d_mod
  use morton_mod
  use morton2d_mod
  use mrgrnk_mod
  use stack_mod
  use stack3_mod
  use rbtree_fd4_block_mod
  use timing_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH

  integer(i_k), parameter, public :: FD4_PART_SFCD_WGTX_MASTER = 1
  integer(i_k), parameter, public :: FD4_PART_SFCD_WGTX_COLL = 2
  integer(i_k), parameter, public :: FD4_PART_SFCD_SCATTER_ALLGATHER = 101
  integer(i_k), parameter, public :: FD4_PART_SFCD_ALLGATHER_BCAST = 102
  
  ! public methods
  public :: fd4_part_sfcd_weight_simple
#ifdef FD4_SFC_DISTR
  public :: fd4_part_sfcd_weight
#endif


#ifdef FD4_SFC_DISTR_VERBOSE  
  integer, parameter :: sfcdbg = 9000
#endif

#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
    integer :: ferr
    ! Declare the interface for POSIX fsync function
!    interface
!      function fsync (fd) bind(c,name="fsync")
!      use iso_c_binding, only: c_int
!        integer(c_int), value :: fd
!        integer(c_int) :: fsync
!      end function fsync
!    end interface
#endif
  
contains  

  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'
  include 'util/stack3.incf'
  include 'util/math.incf'
  include 'framework/fd4_part_sfc.incf'


  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! This version only works in non-adaptive block mode and when the SFC type is never changed.
  !!
  !! Uses simple 1D partitioning heuristic.
  !!
  !! *TODO*:
  !! - update_localweight should be a parameter (in fd4_domain_lb_params)
  !!
  subroutine fd4_part_sfcd_weight_simple(domain, local_blocks, sfctype, blkmap_update, use_h2, err)
    type(fd4_domain), intent(inout) :: domain               !! the domain
    type(stack3)    , intent(inout) :: local_blocks         !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: sfctype              !! space-filling curve to use
    logical         , intent(in)    :: blkmap_update        !! update the full block map
    logical         , intent(in)    :: use_h2               !! use H2 heuristic instead of H1 (Miguet & Pierson)
    integer(i_k)    , intent(out)   :: err                  !! error status: 0...ok
    target :: domain

    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, crank, pstart, pend, oldp(2)
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    type(fd4_mpi), pointer :: mpi
    type(part_sfc_iter) :: sfciter
    integer(i8k) :: timing_us0, timing_us0mpi, timing_us1
    
    type(fd4_domain_balance), pointer :: bal
    real(r8k) :: sumweight, bottleneck, sumload
    integer(i_k) :: part0, part1, part, part0_new, guess
    type(stack) :: borders_found
    integer :: request, status(MPI_STATUS_SIZE)
    integer(i_k) :: border, h2_add_block
    logical :: exact_part_match
    
    logical, parameter :: update_localweight = .true.
    real(r8k) :: localweight_ps(1:2), localweight_ps_send

    err = -1
    
    FD4_L1_ASSERT(domain%varinfo%vthres_mode,'not allowed in adaptive block mode')
    if(domain%blockcount>0 .and. sfctype/=domain%balance%last_sfctype) then
      FD4_L1_ERROR('cannot change SFC type')
    end if
    if(domain%blocks%length>0 .and. domain%balance%sfc_pstart<0) then
      FD4_L1_ERROR('domain%balance%sfc_pstart not set')
    end if

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    nblocks = PRODUCT(domain%blnum(1:3))
   
    if(.not. associated(domain%balance%partition)) then
      allocate(domain%balance%partition(0:domain%mpi%nprocs), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate domain%balance%partition')
      err = -1
    end if
    partition => bal%partition

    bal%last_sfctype = sfctype

    if(domain%blockcount == nblocks) then
      !! all blocks already allocated

      !! 1. parallel prefix sum of block weights
      !!
      
      if(domain%stat_enabled) call gettime_ofday(timing_us0)
      
      call fd4_part_sfc_prefixsum_parallel(domain, sfctype, err, opt_sumweight = sumweight, opt_consistent = .true.)
      FD4_L1_ASSERT(err/=0,'fd4_part_sfc_prefixsum_parallel failed')

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_pfxs_us) = domain%stat%values(domain%balance%stat_pfxs_us) +(timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

      !! 2. get the partition borders on the SFC which reside in local part of the block weight vector
      !!

      bottleneck = sumweight / nprocs
      exact_part_match = .false.

      ! first coarse-grained partition to scan for
      part0 = CEILING(bal%sfcd_blkweights_ps(0) / bottleneck)
      part0 = MAX(1,part0)

      ! last coarse-grained partition to scan for
      part1 = FLOOR(bal%sfcd_blkweights_ps(domain%blocks%length) / bottleneck)
      part1 = part1 + 1
      part1 = MIN(nprocs-1,part1)

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'scan ', part0, part1
      write(sfcdbg+rank,*) 'scan ', bal%sfcd_blkweights_ps(0), bal%sfcd_blkweights_ps(domain%blocks%length)
      write(sfcdbg+rank,'(A,E50.40)') 'scan ', bal%sfcd_blkweights_ps(0) / bottleneck
      write(sfcdbg+rank,'(A,E50.40)') 'scan ', bal%sfcd_blkweights_ps(domain%blocks%length) / bottleneck
#endif

      FD4_L1_ASSERT(part1-MAX(1,part0-1)<0,'part bug')

      ! search with H1 for partitions
      part0_new = HUGE(part0_new)
      pstart = 0
      guess = 0
      h2_add_block = 0
      call stack_create(borders_found, part1-part0+2)

      do part = MAX(1,part0-1), part1
        sumload = part * bottleneck
        ! pend = last block of current partition
        pend = fd4_part_1d_scan_bin(domain%blocks%length, bal%sfcd_blkweights_ps, pstart, domain%blocks%length+1, sumload)
        !pend = fd4_part_1d_scan(domain%blocks%length, bal%sfcd_blkweights_ps, pstart, sumload, guess)
        if(use_h2) then
          h2_add_block = 0
          if(pend < domain%blocks%length) then
            if(bal%sfcd_blkweights_ps(pend+1) - sumload < sumload - bal%sfcd_blkweights_ps(pend)) h2_add_block = 1
          end if
        end if
#ifdef FD4_SFC_DISTR_VERBOSE
        write(sfcdbg+rank,*) 'scan for: ', part, sumload, '->', pend+bal%sfc_pstart-1, &
                           sumload>=bal%sfcd_blkweights_ps(pend), &
                           pend==0,sumload==bal%sfcd_blkweights_ps(pend),pend<domain%blocks%length
#endif
        if(sumload>=bal%sfcd_blkweights_ps(pend)) then
          if(.not. exact_part_match .and. pend==0 .and. sumload==bal%sfcd_blkweights_ps(pend)) then
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) ' exact part match', part, sumload, '->', pend+bal%sfc_pstart-1
#endif
            ! if the previous rank has the same match at its last block
            ! discard this match
            exact_part_match = .true.
            part0_new = part+1
          elseif(pend<domain%blocks%length .or. sumload==bal%sfcd_blkweights_ps(pend)) then
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) ' valid match     ', part, sumload, '->', pend+bal%sfc_pstart-1
#endif
            call stack_push_fast(borders_found, pend+bal%sfc_pstart-1+h2_add_block)
            part0_new = MIN(part, part0_new)
          end if
        end if
        pstart = pend
      end do

      if(part0_new<HUGE(part0_new)) part0 = part0_new
      part1 = part0 + borders_found%pos - 1

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'scan update', part0, part1
      write(sfcdbg+rank,*) 'exact_part_match  ', exact_part_match
#endif


      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
        timing_us0mpi = timing_us1
      end if


      !! 3. tell all ranks the borders of their own partition

      ! reset the stack of requests
      call stack_create(mpi%requests, (part1-part0+1))

      ! loop over borders and send them to resp. rank
      ib = 1
      border = -1
      do part = part0,part1  
        ! send lower partition border to rank == part
        if(part==rank) then
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'copy lower border: rank ',part
#endif
          ! do not send msg to myself, do local copy
          border = borders_found%a(ib) + 1
        else
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'send lower border: rank ',part
#endif
          ! send msg
          call MPI_Isend(borders_found%a(ib), 1, MPI_I_K,  part, 10, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call stack_push_fast(mpi%requests, request)
        end if
        ib = ib + 1
      end do

      ! receive partition borders
      if(border<0) then
        if(rank>0) then
          ! receive lower border
          call MPI_Recv(border, 1, MPI_I_K, MPI_ANY_SOURCE, 10, mpi%comm, status, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
          ! increase by one since the upper bound of previous part is sent
          border = border + 1
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'recv lower border: rank ',status(MPI_SOURCE)
#endif
        else
          border = 1
        end if
      end if

      if(mpi%requests%pos>0) then
        call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
      end if    


      pstart = border


      !! 6: distribute partition vector over all ranks
      !!

      ! save old local partition
      oldp(1) = bal%sfc_pstart
      oldp(2) = oldp(1) + domain%blocks%length - 1

      ! save old partition vector
      if(.not. blkmap_update) then
        if(.not. associated(bal%partition_old)) then
          allocate(bal%partition_old(0:domain%mpi%nprocs), stat=err)
          FD4_L1_ASSERT(err/=0,'allocate bal%partition_old')
          err = -1
        end if
        bal%partition => bal%partition_old
        bal%partition_old => partition
        partition => bal%partition
      end if

      call MPI_Allgather(pstart, 1, MPI_I_K, partition(0), 1, MPI_I_K, mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Allgather failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        domain%stat%values(bal%stat_dspv_us) = domain%stat%values(bal%stat_dspv_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
        timing_us0mpi = timing_us1
      end if
      partition(nprocs) = nblocks + 1

      pend   = partition(rank+1)-1

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) ' partition:', partition
#endif

      if(update_localweight) then

        !! do some extra communication to update domain%localweight

        call stack_create(mpi%requests, (part1-part0+1)+1 )
      
        ! loop over borders and send lower weight prefix sum to resp. rank
        ib = 1
        localweight_ps(:) = -1.0_r8k
        do part = part0,part1
          localweight_ps_send = bal%sfcd_blkweights_ps( borders_found%a(ib) + 1 - bal%sfc_pstart )
          if(part==rank) then
            ! do not send msg to myself, do local copy
            localweight_ps(1) = localweight_ps_send
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'copy lower border:  ',part, borders_found%a(ib), bal%sfc_pstart, localweight_ps_send
#endif
          else
            call MPI_Isend(localweight_ps_send, 1, MPI_R8K, part, 30, mpi%comm, request, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
            call stack_push_fast(mpi%requests, request)
          end if
          ib = ib + 1
        end do

        ! receive weight
        if(localweight_ps(1) < 0.0_r8k) then
          if(rank>0) then
            call MPI_Recv(localweight_ps(1), 1, MPI_R8K, MPI_ANY_SOURCE, 30, mpi%comm, MPI_STATUS_IGNORE, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
          else
            localweight_ps(1) = 0.0_r8k
          end if
        end if
        
        ! send weight to previous rank
        if(rank>0) then
          call MPI_Isend(localweight_ps(1), 1, MPI_R8K, rank-1, 40, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call stack_push_fast(mpi%requests, request)
        end if
        
        ! receive my upper prefix sum weight from next rank (its lower prefix sum weight)
        if(rank<nprocs-1) then
          call MPI_Recv(localweight_ps(2), 1, MPI_R8K, rank+1, 40, mpi%comm, MPI_STATUS_IGNORE, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
        else
          localweight_ps(2) = bal%sfcd_blkweights_ps(domain%blocks%length)
        end if
        
#ifdef FD4_SFC_DISTR_VERBOSE
        write(sfcdbg+rank,*) ' localweight_ps:', localweight_ps
        write(sfcdbg+rank,*) ' localweight:', localweight_ps(2) - localweight_ps(1)
#endif
        
        bal%localweight = localweight_ps(2) - localweight_ps(1)

        if(mpi%requests%pos>0) then
          call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
        end if    

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        end if

      else  ! update_localweight
  
        bal%localweight = 0.0_r8k
    
      end if

      bal%sfc_pstart = pstart

    elseif(domain%blockcount == 0) then
      !! no blocks allocated yet
      
      if(domain%stat_enabled) call gettime_ofday(timing_us0)

      !! create an initial partitioning
      !! TODO: consider weights (i.e. block sizes) here

      do crank=0,nprocs
        partition(crank) = INT( (INT(nblocks,i8k) * INT(crank,i8k)) / INT(nprocs,i8k) + 1_i8k, i_k )
      end do
      
      pstart = partition(rank)
      pend   = partition(rank+1)-1
      bal%sfc_pstart = pstart
      bal%localweight = 0.0_r8k
      oldp(1:2) = -1

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'initial partition: ', partition
#endif

    else
      !! adaptive block mode
      FD4_L1_ERROR('not allowed in adaptive block mode')
    end if


    !! finally apply partition to FD4
    !!

    err = -1

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now

    ! allocate local list of blocks
    if(pstart<=pend) then
      call stack3_create(local_blocks, pend - pstart + 1)
      local_blocks%pos = pend - pstart + 1
    else
      local_blocks%pos = 0
    end if

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'pstart ',pstart
      write(sfcdbg+rank,*) 'pend   ',pend
      write(sfcdbg+rank,*) 'local_blocks%pos   ',local_blocks%pos
      write(sfcdbg+rank,*) ''
#endif    

    call fd4_part_sfc_iter_init(sfciter, domain, sfctype)

    ! set local list of blocks
    if(pstart<=pend) then
      call fd4_part_sfc_iter_set(sfciter, pstart)
      do while(sfciter%idx>0 .and. sfciter%idx<=pend)
        local_blocks%a(1:3,sfciter%idx-pstart+1) = sfciter%p(1:3)
        call fd4_part_sfc_iter_next(sfciter)
      end do
    end if

    bal%blkmap_needs_update = .true.
    
    ! here is the actual blkmap update
    if(blkmap_update) then
      ! full update
      call fd4_part_sfc_update_blkmap(domain)
    else
      ! update only the necessary blocks
      call fd4_part_sfc_update_blkmap_loc(domain, sfciter, oldp, pstart, pend, local_blocks, err)
      FD4_L1_ASSERT(err/=0,'fd4_part_sfc_update_blkmap_loc failed')
      err = -1
    end if

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(domain%balance%stat_blkm_us) = domain%stat%values(domain%balance%stat_blkm_us) + (timing_us1-timing_us0)
      ! fill statistics
      domain%stat%sfcbnd(1) = MIN(domain%stat%sfcbnd(1), pstart)
      domain%stat%sfcbnd(2) = MAX(domain%stat%sfcbnd(2), pend)
    end if

#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks%pos, domain%balance%localweight
#endif
  
    err = 0

  end subroutine fd4_part_sfcd_weight_simple

#ifdef FD4_SFC_DISTR

  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! This version only works in non-adaptive block mode and when the SFC type is never changed.
  !!
  !! - $wgtxmeth may be $FD4_PART_SFCD_WGTX_MASTER or $FD4_PART_SFCD_WGTX_COLL
  !!   ($FD4_PART_SFCD_WGTX_COLL is recomended).
  !! - $wgtxmeth = $FD4_PART_SFCD_WGTX_MASTER implies $p1d_parallel == true
  !! - $method = $FD4_PART_SFC_PARALLEL implies $p1d_parallel == true
  !! - if $p1d_parallel is false, domain%balancec%localweight is not set
  !! - $dspv_meth can be: $FD4_PART_SFCD_SCATTER_ALLGATHER or $FD4_PART_SFCD_ALLGATHER_BCAST
  !!
  !! *EXPERIMENTAL DISTRIBUTED/HIERARCHICAL VERSION*:
  !!  - ranks are splitted into equal-sized groups of size $groupsz
  !!  - a coarse-grained partitioning for the groups is calculated with a fast and parallel heuristic
  !!  - for each group a high-quality method is used to create the final partition for each rank
  !!
  subroutine fd4_part_sfcd_weight(domain, local_blocks, method, p1d_parallel, groupsz, sfctype, &
                                  binsteps, accuracy, blkmap_update, wgtxmeth, use_h2, dspv_meth, err)
    type(fd4_domain), intent(inout) :: domain               !! the domain
    type(stack3)    , intent(inout) :: local_blocks         !! stack3 of new local partition (differs from rank to rank)
    integer(i_k)    , intent(in)    :: method               !! 1D partitioning method to use
    logical         , intent(in)    :: p1d_parallel         !! perform high-quality 1D partitioning in parallel?
    integer(i_k)    , intent(in)    :: groupsz              !! size of groups
    integer(i_k)    , intent(in)    :: sfctype              !! space-filling curve to use
    integer(i_k)    , intent(in)    :: binsteps             !! max. number of steps for binary search methods
    !! target ratio from best possible load balance for binary steps estimation, *must be <= 1.0*
    real(r8k)       , intent(in)    :: accuracy
    logical         , intent(in)    :: blkmap_update        !! update the full block map
    integer(i_k)    , intent(in)    :: wgtxmeth             !! method for exchanging block weights
    logical         , intent(in)    :: use_h2               !! use H2 heuristic instead of H1 (Miguet & Pierson)
    integer         , intent(in)    :: dspv_meth            !! method of partition vector distribution when p1d_parallel is false
    integer(i_k)    , intent(out)   :: err                  !! error status: 0...ok
    target :: domain

    integer :: rank, nprocs, nblocks
    integer(i_k) :: ib, crank, pstart, pend, oldp(2), reqsteps
    real(r8k) :: maxweight, maxload, eps
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    type(fd4_mpi), pointer :: mpi
    type(part_sfc_iter) :: sfciter
    integer(i8k) :: mpi_time, timing_us0, timing_us0mpi, timing_us1
    
    type(fd4_domain_balance), pointer :: bal
    real(r8k) :: sumweight, bottleneck, sumload
    integer(i_k) :: cgpart0, cgpart1, cgpart, cgpart0_new, guess, ngroups
    type(stack) :: cgborders_found
    integer :: request, status(MPI_STATUS_SIZE)
    integer(i_k) :: cgborders(4), ncgblocks, h2_add_block
    integer(i_k) :: local_copy_start, local_copy_type
    logical :: exact_part_match, p1dpara
#ifndef FD4_SFCD_H2_FAST
    logical :: h2_add_block_at_end
#endif
#ifdef FD4_SFC_DISTR_DEBUG
    integer(i_k) :: emergency_abort
#endif

    err = -1
    
    FD4_L1_ASSERT(domain%varinfo%vthres_mode,'not allowed in adaptive block mode')
    if(domain%blockcount>0 .and. sfctype/=domain%balance%last_sfctype) then
      FD4_L1_ERROR('cannot change SFC type')
    end if
    if(domain%blocks%length>0 .and. domain%balance%sfc_pstart<0) then
      FD4_L1_ERROR('domain%balance%sfc_pstart not set')
    end if

    p1dpara = p1d_parallel .or. (wgtxmeth == FD4_PART_SFCD_WGTX_MASTER) .or. (method == FD4_PART_SFC_PARALLEL)

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    nblocks = PRODUCT(domain%blnum(1:3))
   
    if(.not. associated(domain%balance%partition)) then
      allocate(domain%balance%partition(0:domain%mpi%nprocs), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate domain%balance%partition')
      err = -1
    end if
    partition => domain%balance%partition

    bal%last_sfctype = sfctype
    
#ifdef FD4_SFC_DISTR_VERBOSE
    write(sfcdbg+rank,*) '** ENTER fd4_part_sfcd_weight ** domain%blocks%length = ', domain%blocks%length
#endif

    !! create the group communicator if first call
    if(bal%sfcd_ngroups==0) then

      if(domain%stat_enabled) call gettime_ofday(timing_us0)
      
      FD4_L1_ASSERT(groupsz<=1,'groupsz must be > 1')
      FD4_L1_ASSERT(groupsz>nprocs/2,'groupsz must be <= nprocs/2')
      bal%sfcd_groupsz = groupsz
      bal%sfcd_ngroups = nprocs / groupsz
      if( bal%sfcd_groupsz * bal%sfcd_ngroups /= nprocs ) then
        bal%sfcd_groupsz = 0
        bal%sfcd_ngroups = 0
        FD4_L1_ERROR('groupsz must be divisor of nprocs')
      end if
      
      bal%sfcd_mygroup = rank / bal%sfcd_groupsz
      call MPI_Comm_split(mpi%comm, bal%sfcd_mygroup, rank, bal%sfcd_mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Comm_split failed')
      call MPI_Comm_rank(bal%sfcd_mpi%comm, bal%sfcd_mpi%rank, err)
      call MPI_Comm_size(bal%sfcd_mpi%comm, bal%sfcd_mpi%nprocs, err)
      err = -1
      FD4_L1_ASSERT(bal%sfcd_mpi%nprocs/=groupsz,'comm has wrong size')
#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) bal%sfcd_mygroup, bal%sfcd_mpi%rank
#endif
      bal%sfcd_mpi%comm_gc  = bal%sfcd_mpi%comm
      bal%sfcd_mpi%comm_cpl = bal%sfcd_mpi%comm
      bal%sfcd_mpi%comm_mig = bal%sfcd_mpi%comm
      bal%sfcd_mpi%lognprocs = log2i(bal%sfcd_mpi%nprocs)
      
      if(dspv_meth==FD4_PART_SFCD_ALLGATHER_BCAST) then
        !! create communicator containing all group masters
        
        ib = MPI_UNDEFINED
        if(bal%sfcd_mpi%rank==0) ib = 1
        call MPI_Comm_split(mpi%comm, ib, rank, bal%comm_sfcd_master, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Comm_split failed')
      
      end if

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_init_us) = domain%stat%values(domain%balance%stat_init_us) + (timing_us1-timing_us0)
      end if

    elseif(bal%sfcd_groupsz/=groupsz) then
      FD4_L1_ERROR('changing group size not implemented')
    end if

    ngroups = bal%sfcd_ngroups

    if(domain%blockcount == nblocks) then
      !! all blocks already allocated

      !! 1. parallel prefix sum of block weights
      !!
      
      if(domain%stat_enabled) call gettime_ofday(timing_us0)
      
      call fd4_part_sfc_prefixsum_parallel(domain, sfctype, err, opt_sumweight = sumweight, opt_consistent = .true.)
      FD4_L1_ASSERT(err/=0,'fd4_part_sfc_prefixsum_parallel failed')
      err = -1

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_pfxs_us) = domain%stat%values(domain%balance%stat_pfxs_us) +(timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

      !! 2. get the coarse-grained partition borders on the SFC which
      !!    reside in local part of the block weight vector

      bottleneck = sumweight / ngroups
      exact_part_match = .false.

      ! first coarse-grained partition to scan for
      cgpart0 = CEILING(bal%sfcd_blkweights_ps(0) / bottleneck)
      cgpart0 = MAX(1,cgpart0)
      cgpart0 = MIN(ngroups,cgpart0)

      ! last coarse-grained partition to scan for
      cgpart1 = FLOOR(bal%sfcd_blkweights_ps(domain%blocks%length) / bottleneck)
      cgpart1 = cgpart1 + 1
      cgpart1 = MIN(ngroups-1,cgpart1)

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'coarse scan ', cgpart0, cgpart1
      write(sfcdbg+rank,*) 'coarse scan ', bal%sfcd_blkweights_ps(0), bal%sfcd_blkweights_ps(domain%blocks%length)
      write(sfcdbg+rank,'(A,E50.40)') 'coarse scan ', bal%sfcd_blkweights_ps(0) / bottleneck
      write(sfcdbg+rank,'(A,E50.40)') 'coarse scan ', bal%sfcd_blkweights_ps(domain%blocks%length) / bottleneck
#endif

      FD4_L1_ASSERT(cgpart1<MAX(1,cgpart0-1),'cgpart '//i2s(cgpart0))

      ! search with H1 for coarse-grained partitions
      cgpart0_new = HUGE(cgpart0_new)
      pstart = 0
      guess = 0
      h2_add_block = 0
#ifndef FD4_SFCD_H2_FAST
      h2_add_block_at_end = .false.
#endif
      call stack_create(cgborders_found, cgpart1-cgpart0+2)

      do cgpart = MAX(1,cgpart0-1), cgpart1
        sumload = cgpart * bottleneck
        ! pend = last block of current partition
        pend = fd4_part_1d_scan_bin(domain%blocks%length, bal%sfcd_blkweights_ps, pstart, domain%blocks%length+1, sumload)
        !pend = fd4_part_1d_scan(domain%blocks%length, bal%sfcd_blkweights_ps, pstart, sumload, guess)
        if(use_h2) then
          h2_add_block = 0
          !
          ! CAUTION: "domain%blocks%length-1" is actually wrong in this condition. It should be "domain%blocks%length",
          ! but this workaround avoids coarse partition matches at the very end the the own blocks without notice of
          ! the next rank. This would lead to deadlocks in the weight exchange. Of course, one can solve this issue...
          ! (i.e. get sfcd_blkweights_ps(-1) from previous rank and check the same condition)
          ! This workaround reduces the H2 search accuracy at partition borders, but should still lead to better
          ! results than H1 search.
          !
          if(pend < domain%blocks%length-1) then
            if(bal%sfcd_blkweights_ps(pend+1) - sumload < sumload - bal%sfcd_blkweights_ps(pend)) h2_add_block = 1
#ifndef FD4_SFCD_H2_FAST
          else if(pend == domain%blocks%length-1) then
            if(bal%sfcd_blkweights_ps(pend+1) - sumload < sumload - bal%sfcd_blkweights_ps(pend)) then
              h2_add_block = 1
              h2_add_block_at_end = .true.
            end if
#endif
          end if
        end if
#ifdef FD4_SFC_DISTR_VERBOSE
        write(sfcdbg+rank,*) 'coarse scan for: ', cgpart, sumload, '->', pend+bal%sfc_pstart-1, &
                           sumload>=bal%sfcd_blkweights_ps(pend), &
                           pend==0,sumload==bal%sfcd_blkweights_ps(pend),pend<domain%blocks%length
#endif
        if(sumload>=bal%sfcd_blkweights_ps(pend)) then
          if(.not. exact_part_match .and. pend==0 .and. sumload==bal%sfcd_blkweights_ps(pend)) then
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) ' exact part match', cgpart, sumload, '->', pend+bal%sfc_pstart-1
#endif
            ! if the previous rank has the same match at its last block
            ! discard this match
            exact_part_match = .true.
            cgpart0_new = cgpart+1
          elseif(pend<domain%blocks%length .or. sumload==bal%sfcd_blkweights_ps(pend)) then
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) ' valid match     ', cgpart, sumload, '->', pend+bal%sfc_pstart-1, '  +', h2_add_block
#endif
            call stack_push_fast(cgborders_found, pend+bal%sfc_pstart-1+h2_add_block)
            cgpart0_new = MIN(cgpart, cgpart0_new)
          end if
        end if
        pstart = pend
      end do

      if(cgpart0_new<HUGE(cgpart0_new)) cgpart0 = cgpart0_new
      cgpart1 = cgpart0 + cgborders_found%pos - 1

#ifndef FD4_SFCD_H2_FAST
      ! tell rank+1 whether I had a match at the last block due to H2
      ! TODO: if rank+1 does not own blocks, send to smallest rank > my rank owning blocks in current partitioning
      if(use_h2 .and. domain%blocks%length>0) then
        
        if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

        if(rank<nprocs-1) then
          crank = rank + 1
          ! if crank's current partition is empty, we search for the next rank and so on
          do while(partition(crank+1)==partition(crank))
            crank = crank + 1
            if(crank>=nprocs) exit
          end do
          if(crank<nprocs) then
            call MPI_Bsend(h2_add_block_at_end, 1, MPI_LOGICAL, crank, 50, mpi%comm, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Bsend failed')
            err = -1
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'send h2_add_block_at_end to ',crank,': ',h2_add_block_at_end
#endif
          end if
        end if
        
        if(rank>0) then
          crank = rank - 1
          ! if crank's current partition is empty, we search for the previous rank and so on
          do while(partition(crank+1)==partition(crank))
            crank = crank - 1
            if(crank<0) exit
          end do
          if(crank>=0) then
            call MPI_Recv(h2_add_block_at_end, 1, MPI_LOGICAL, crank, 50, mpi%comm, MPI_STATUS_IGNORE, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
            err = -1
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'recv h2_add_block_at_end from ',crank,': ',h2_add_block_at_end
#endif
            exact_part_match = ( exact_part_match .or. h2_add_block_at_end )
          end if
        end if
      
      end if
#endif

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'coarse scan update', cgpart0, cgpart1
      write(sfcdbg+rank,*) 'exact_part_match  ', exact_part_match
#endif


      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1dc_us) = domain%stat%values(bal%stat_p1dc_us) + (timing_us1-timing_us0)
#ifndef FD4_SFCD_H2_FAST
        if(use_h2 .and. domain%blocks%length>0) then
          domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        end if
#endif
        timing_us0 = timing_us1
        timing_us0mpi = timing_us1
      end if

#ifdef FD4_SFC_DISTR_DEBUG
      call MPI_Allreduce(cgborders_found%pos, emergency_abort, 1, MPI_I_K, MPI_SUM, mpi%comm, err)
      if(emergency_abort/=ngroups-1) then
        if(rank==0) then
          write(sfcdbg+rank,*) '!! ',emergency_abort,' borders found !!'
          write(*,*) '!! ',emergency_abort,' borders found !!'
        end if
        call mpi_finalize(err)
        stop
      end if
#endif


      !! 3. tell all groups the borders of their coarse-grained partition
      !!    (border finder sends to groups master, master bcasts to all groups members)

      ! reset the stack of requests
      call stack_create(mpi%requests, 2*(cgpart1-cgpart0+1))

      ! loop over borders and send them to resp. group master
      ib = 1
      local_copy_start = 0
      local_copy_type = 0
      cgborders(:) = -1
      do cgpart = cgpart0,cgpart1  
        ! send partition border to ranks cgpart*groupsz and (cgpart-1)*groupsz
        ! with a tag to identify lower and upper borders
        ! lower border for cgpart, tag = 10
        if(cgpart*groupsz==rank) then
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'copy lower border: group/rank ',cgpart, cgpart*groupsz
#endif
          ! do not send msg to myself, do local copy
          cgborders(1) = cgborders_found%a(ib) + 1
          cgborders(3) = rank
        else
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'send lower border: group/rank ',cgpart, cgpart*groupsz
#endif
          ! send msg
          call MPI_Isend(cgborders_found%a(ib), 1, MPI_I_K,  cgpart   *groupsz, 10, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call stack_push_fast(mpi%requests, request)
        end if
        ! upper border for cgpart-1, tag = 11
        if((cgpart-1)*groupsz==rank) then
#ifdef FD4_SFC_DISTR_VERBOSE
           write(sfcdbg+rank,*) 'copy upper border: group/rank ',cgpart-1, (cgpart-1)*groupsz
#endif
          ! do not send msg to myself, do local copy
          cgborders(2) = cgborders_found%a(ib)
          cgborders(4) = rank
        else
#ifdef FD4_SFC_DISTR_VERBOSE
           write(sfcdbg+rank,*) 'send upper border: group/rank ',cgpart-1, (cgpart-1)*groupsz
#endif
          call MPI_Isend(cgborders_found%a(ib), 1, MPI_I_K, (cgpart-1)*groupsz, 11, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call stack_push_fast(mpi%requests, request)
        end if
        ib = ib + 1
      end do

      ! group masters: receive partition borders
      if(bal%sfcd_mpi%rank == 0) then
      
        ! lower border
        if(cgborders(1)<0) then
          if(bal%sfcd_mygroup>0) then
            ! receive lower border, tag = 10
            call MPI_Recv(cgborders(1), 1, MPI_I_K, MPI_ANY_SOURCE, 10, mpi%comm, status, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
            ! increase by one since the upper bound of previous cgpart is sent
            cgborders(1) = cgborders(1) + 1
            cgborders(3) = status(MPI_SOURCE)
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'recv lower border: rank ',status(MPI_SOURCE)
#endif
          else
            cgborders(1) = 1
            cgborders(3) = 0
          end if
        end if
        
        ! find the rank who owns the lower border (should be cgborders(3) or a bit higher)
        do while(partition(cgborders(3)+1)<=cgborders(1))
          cgborders(3) = cgborders(3) + 1
          if(cgborders(3)>=nprocs) exit  ! empty coarse partition
        end do

        ! upper border
        if(cgborders(2)<0) then
          if(bal%sfcd_mygroup<ngroups-1) then
            ! receive upper border, tag = 11
            call MPI_Recv(cgborders(2), 1, MPI_I_K, MPI_ANY_SOURCE, 11, mpi%comm, status, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
            cgborders(4) = status(MPI_SOURCE)
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'recv upper border: rank ',status(MPI_SOURCE)
#endif
          else
            cgborders(2) = nblocks
            cgborders(4) = nprocs - 1
          end if
        end if

        ! find the rank who owns the upper border (should be cgborders(4) or a bit lower)
        do while(partition(cgborders(4))>cgborders(2))
          cgborders(4) = cgborders(4) - 1
          if(cgborders(4)<0) exit  ! empty coarse partition
        end do

      end if

      ! bcast partition borders to groups members
      call MPI_Bcast(cgborders(1), 4, MPI_I_K, 0, bal%sfcd_mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Bcast failed')

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'coarse borders: ',cgborders(1:2)
      write(sfcdbg+rank,*) 'coarse border ranks: ',cgborders(3:4)
#endif

      ! number of blocks in coarse grained partition of local group
      ncgblocks = cgborders(2) - cgborders(1) + 1
      
      if(ncgblocks==0 .and. bal%sfcd_mpi%rank==0) then
        FD4_L2_INFO('empty coarse partition')
      end if
      
      FD4_L1_ASSERT(cgborders(3)>=nprocs.and.ncgblocks>0,'lo cgrank')
      FD4_L1_ASSERT(cgborders(4)<0.and.ncgblocks>0,'up cgrank')

      if(mpi%requests%pos>0) then
        call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
      end if    

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        domain%stat%values(bal%stat_bcbr_us) = domain%stat%values(bal%stat_bcbr_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

      err = -1


      !! 4. exchange block weights
      !!
      
      if(wgtxmeth == FD4_PART_SFCD_WGTX_MASTER) then

        call fd4_part_sfcd_exch_weights(domain, cgpart0, cgborders_found, cgborders, exact_part_match, err)
        FD4_L1_ASSERT(err/=0,'fd4_part_sfcd_exch_weights failed')

      else
      
        call fd4_part_sfcd_exch_weights_coll(domain, cgpart0, cgborders_found, cgborders, exact_part_match, p1dpara, err)
        FD4_L1_ASSERT(err/=0,'fd4_part_sfcd_exch_weights_coll failed')

      end if

      blkweights_ps => domain%balance%blkweights_ps

      call stack_delete(cgborders_found)

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(domain%balance%stat_exwgt_us) = domain%stat%values(domain%balance%stat_exwgt_us) &
                                                         + (timing_us1-timing_us0)
        timing_us0 = timing_us1
        timing_us0mpi = timing_us1
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      if(ncgblocks>0) then
        do ib=0,ncgblocks
          write(sfcdbg+rank,*) ib, ib+cgborders(1)-1, blkweights_ps(ib)
        end do
      end if
#endif


      !! 5. Create high-quality partitioning within group
      !!

      ! save old local partition
      oldp(1) = bal%sfc_pstart
      oldp(2) = oldp(1) + domain%blocks%length - 1

      ! save old partition vector
      if(.not. blkmap_update) then
        if(.not. associated(bal%partition_old)) then
          allocate(bal%partition_old(0:domain%mpi%nprocs), stat=err)
          FD4_L1_ASSERT(err/=0,'allocate bal%partition_old')
          err = -1
        end if
        bal%partition => bal%partition_old
        bal%partition_old => partition
        partition => bal%partition
      end if

      ! prefix sum of group's blocks:  blkweights_ps(0) ... blkweights_ps(ncgblocks)

      if(p1dpara .or. bal%sfcd_mpi%rank==0) then
      
        if(ncgblocks>0) then

          ! get max block weight in group
          maxweight=blkweights_ps(1)-blkweights_ps(0)
          do ib=2,ncgblocks
            maxweight = MAX(blkweights_ps(ib)-blkweights_ps(ib-1), maxweight)
          end do

          if(method==FD4_PART_SFC_BINSRCH .or. method==FD4_PART_SFC_PARALLEL .or. method==FD4_PART_SFC_EBS) then
            !accuracy=MAX(MIN(accuracy,1.0_r8k),0.1_r8k)
            !binsteps=MAX(MIN(binsteps,10000),1)
            ! compute the bottleneck error value which is necessary to get
            !  (accuracy * optimal_balance)
            ! as the resulting loadbalance of the 1d partitioning
            !  (eps is 0 if accuracy is 1.0: in this case the methods is exact if binsteps is large enough)
            eps = (1.0_r8k-accuracy) * blkweights_ps(ncgblocks) / (accuracy*groupsz)
            ! this would be the necessary number of binary search steps for the given accuracy:
            !steps = log2i( CEILING( maxweight / eps, i_k ) )
          end if

          !! call partitioning routine

          if(method==FD4_PART_SFC_BINSRCH) then
            ! good serial method
            call fd4_part_1d_binsrch(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), maxweight, eps, binsteps, &
                                             partition, maxload, reqsteps)
            if(domain%stat_enabled) then
              domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
            end if
            !
          elseif(method==FD4_PART_SFC_EBS) then
            ! exact bisection from pinar
            call fd4_part_1d_binsrch_ebs(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), maxweight, eps, binsteps, .true., &
                                         partition, maxload, reqsteps)
            if(domain%stat_enabled) then
              domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
            end if
            !
          elseif(method==FD4_PART_SFC_PARALLEL) then    
            ! parallel method
            call fd4_part_1d_parallel(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), maxweight, eps, binsteps, &
                                              bal%sfcd_mpi, partition, maxload, reqsteps, mpi_time, err)
            FD4_L1_ASSERT(err/=0,'fd4_part_1d_parallel failed')
            err = -1
            if(domain%stat_enabled) then
              domain%stat%values(bal%stat_binsteps) = MAX(domain%stat%values(bal%stat_binsteps), INT(reqsteps,i8k))
              domain%stat%values(bal%stat_mpi_us)   = domain%stat%values(bal%stat_mpi_us) + mpi_time
            end if
            !
          elseif(method==FD4_PART_SFC_RB) then
            ! recursive bisection
            call fd4_part_1d_simple_rb(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), partition)
            maxload = blkweights_ps(nblocks) ! forces weird value for balance
            !
          elseif(method==FD4_PART_SFC_SIMPLE) then
            ! simple method 1
            call fd4_part_1d_simple_h1(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), partition, maxload)
            !
          else
            ! simple method 2
            call fd4_part_1d_simple_h2(ncgblocks, groupsz, blkweights_ps(0:ncgblocks), partition, maxload)
            !
          end if

        else ! ncgblocks <= 0

          ! no blocks in coarse grained partition (awkward, but valid)
          maxweight = 0.0_r8k
          partition(0:groupsz) = 1

        end if

#ifdef FD4_SFC_DISTR_DEBUG_BALANCE
        if(bal%sfcd_mpi%rank == 0) then
          maxload=0.0_r8k
          do ib=0,groupsz-1
            maxload = MAX(maxload, blkweights_ps(partition(ib+1)-1) - blkweights_ps(partition(ib)-1) )
          end do
          write(8000+bal%sfcd_mygroup,'(I6,I12,I12,F16.4,F16.4,F14.5,F16.4,F14.5)') &
              bal%sfcd_mygroup, cgborders(1:2), blkweights_ps(ncgblocks), maxweight, &
              blkweights_ps(ncgblocks)/(maxweight*ncgblocks), &
              maxload, blkweights_ps(ncgblocks)/(maxload*groupsz)
        end if
#endif

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
          timing_us0 = timing_us1
          timing_us0mpi = timing_us1
        end if

      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      do ib=0,groupsz
        write(sfcdbg+rank,*) 'partition: ',ib+bal%sfcd_mygroup*groupsz, partition(ib)+cgborders(1)-1
      end do
#endif    

      !! 6: distribute partition vector over all ranks
      !!


      if(dspv_meth==FD4_PART_SFCD_ALLGATHER_BCAST) then
        !  use allgather (only masters) + bcast (within group)
        if(bal%comm_sfcd_master/=MPI_COMM_NULL) then
#ifdef FD4_NO_MPI_IN_PLACE
          call fd4_mpi_realloc_array(mpi%recvcnt, bal%sfcd_groupsz, 0, err)
          FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')
          mpi%recvcnt(0:bal%sfcd_groupsz-1) = partition(0:bal%sfcd_groupsz-1) + cgborders(1)-1
          call MPI_Allgather(mpi%recvcnt(0), bal%sfcd_groupsz, MPI_I_K, partition(0), bal%sfcd_groupsz, MPI_I_K, &
                             bal%comm_sfcd_master, err)
#else
          partition(bal%sfcd_groupsz*bal%sfcd_mygroup:bal%sfcd_groupsz*(bal%sfcd_mygroup+1)-1) = &
            partition(0:bal%sfcd_groupsz-1) + cgborders(1)-1
          call MPI_Allgather(MPI_IN_PLACE, 0, 0, partition(0), bal%sfcd_groupsz, MPI_I_K, bal%comm_sfcd_master, err)
#endif
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Allgather failed')
        end if
        call MPI_Bcast(partition(0), nprocs, MPI_I_K, 0, bal%sfcd_mpi%comm, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Bcast failed')
        err = -1
        pstart = partition(rank)
      else ! if(dspv_meth==FD4_PART_SFCD_SCATTER_ALLGATHER) then
        ! use scatter (within group) + allgather (global)
        if(p1dpara) then
          pstart = partition(bal%sfcd_mpi%rank) + cgborders(1)-1
        else
          call MPI_Scatter(partition(0), 1, MPI_I_K, pstart, 1, MPI_I_K, 0, bal%sfcd_mpi%comm, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Scatter failed')
          pstart = pstart + cgborders(1)-1
        end if
        call MPI_Allgather(pstart, 1, MPI_I_K, partition(0), 1, MPI_I_K, mpi%comm, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Allgather failed')
        err = -1
      end if


      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        domain%stat%values(bal%stat_dspv_us) = domain%stat%values(bal%stat_dspv_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

      partition(nprocs) = nblocks + 1

      pend   = partition(rank+1)-1
      bal%sfc_pstart = pstart
      
      if(p1dpara) then
        if(pstart<=pend) then
          bal%localweight = blkweights_ps(pend-cgborders(1)+1) - blkweights_ps(pstart-cgborders(1))
        else
          bal%localweight = 0.0_r8k
        end if
      else
        bal%localweight = -8192.0_r8k
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      if(rank==0) print *,partition
#endif

    elseif(domain%blockcount == 0) then
      !! no blocks allocated yet
      
      if(domain%stat_enabled) call gettime_ofday(timing_us0)

      !! create an initial partitioning
      !! TODO: consider weights (i.e. block sizes) here

      do crank=0,nprocs
        partition(crank) = INT( (INT(nblocks,i8k) * INT(crank,i8k)) / INT(nprocs,i8k) + 1_i8k, i_k )
      end do
      
      pstart = partition(rank)
      pend   = partition(rank+1)-1
      bal%sfc_pstart = pstart
      bal%localweight = 0.0_r8k
      oldp(1:2) = -1

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_p1d_us) = domain%stat%values(bal%stat_p1d_us) + (timing_us1-timing_us0)
        timing_us0 = timing_us1
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      if(rank==0) print *,'initial partition:'
      if(rank==0) print *,partition
#endif

    else
      !! adaptive block mode
      FD4_L1_ERROR('not allowed in adaptive block mode')
    end if


    !! finally apply partition to FD4
    !!

    err = -1

    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now

    ! allocate local list of blocks
    if(pstart<=pend) then
      call stack3_create(local_blocks, pend - pstart + 1)
      local_blocks%pos = pend - pstart + 1
    else
      local_blocks%pos = 0
    end if

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'pstart ',pstart
      write(sfcdbg+rank,*) 'pend   ',pend
      write(sfcdbg+rank,*) 'local_blocks%pos   ',local_blocks%pos
#endif    

    call fd4_part_sfc_iter_init(sfciter, domain, sfctype)

    ! set local list of blocks
    if(pstart<=pend) then
      call fd4_part_sfc_iter_set(sfciter, pstart)
      do while(sfciter%idx>0 .and. sfciter%idx<=pend)
        local_blocks%a(1:3,sfciter%idx-pstart+1) = sfciter%p(1:3)
        call fd4_part_sfc_iter_next(sfciter)
      end do
    end if

    bal%blkmap_needs_update = .true.
    
    ! here is the actual blkmap update
    if(blkmap_update) then
      ! full update
      call fd4_part_sfc_update_blkmap(domain)
    else
      ! update only the necessary blocks
      call fd4_part_sfc_update_blkmap_loc(domain, sfciter, oldp, pstart, pend, local_blocks, err)
      FD4_L1_ASSERT(err/=0,'fd4_part_sfc_update_blkmap_loc failed')
      err = -1
    end if

    if(domain%stat_enabled) then
      call gettime_ofday(timing_us1)
      domain%stat%values(domain%balance%stat_blkm_us) = domain%stat%values(domain%balance%stat_blkm_us) + (timing_us1-timing_us0)
      ! fill statistics
      domain%stat%sfcbnd(1) = MIN(domain%stat%sfcbnd(1), pstart)
      domain%stat%sfcbnd(2) = MAX(domain%stat%sfcbnd(2), pend)
    end if

#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks%pos, domain%balance%localweight
#endif
  
    err = 0

  end subroutine fd4_part_sfcd_weight


  !! Exchange block weights.
  !! This is $FD4_PART_SFCD_WGTX_MASTER.
  !!
  !! Method:
  !! - weights are sent to group master with p2p messages
  !! - master bcasts to group members
  !!
  subroutine fd4_part_sfcd_exch_weights(domain, cgpart0, cgborders_found, cgborders, exact_part_match, err)
    type(fd4_domain), intent(inout) :: domain              !! the domain
    integer(i_k)    , intent(in)    :: cgpart0             !! lower bound for search of local coarse border matches
    type(stack)     , intent(in)    :: cgborders_found     !! stack containing local matches of coarse partition borders
    integer(i_k)    , intent(in)    :: cgborders(4)        !! partitions borders/ranks of own coarse partition
    logical         , intent(in)    :: exact_part_match    !! first local block exactly starts new coarse partition?
    integer(i_k)    , intent(out)   :: err                 !! error status: 0...ok
    target :: domain
    
    integer :: rank, nprocs
    integer(i_k) :: ib, crank
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: timing_us0mpi, timing_us1
    type(fd4_domain_balance), pointer :: bal
    integer(i_k) :: cgpart1, cgpart
    integer :: request, tag
    integer(i_k) :: ncgblocks, start_pos, num_elem, groupsz
    integer(i_k) :: local_copy_start, local_copy_num
    logical :: first_rank_in_cgpart
    ! add this to message tag so that overlap with lower tags not possible
    integer, parameter :: mintag = 20
#ifdef FD4_SFC_DISTR_DEBUG
    integer(i_k) :: emergency_abort
    emergency_abort=0
#endif

    err = -1

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    partition => domain%balance%partition
    groupsz = bal%sfcd_groupsz
    cgpart1 = cgpart0 + cgborders_found%pos - 1
    ncgblocks = cgborders(2) - cgborders(1) + 1
    mpi%requests%pos = 0

    ! send block weights to group masters
    ib = 1
    start_pos = 1
    local_copy_start = 0
    local_copy_num = -1
    ! first_rank_in_cgpart is true before the loop if
    !   this ranks current partition starts with first block of any coarse partition (not only own coarse partition!!)
    first_rank_in_cgpart = (partition(rank)==1 .or. exact_part_match)  ! does not work with H2 coarse search
    !first_rank_in_cgpart = (bal%sfc_pstart==cgborders(1))
#ifdef FD4_SFC_DISTR_VERBOSE
    write(sfcdbg+rank,*) 'send weights: first_rank_in_cgpart ',first_rank_in_cgpart
    write(sfcdbg+rank,*) 'send weights: partition(rank)==1   ',partition(rank)==1
    write(sfcdbg+rank,*) 'send weights: exact_part_match     ',exact_part_match
#endif
    do cgpart = cgpart0-1, cgpart1
      if(cgpart<cgpart1) then
        num_elem = cgborders_found%a(ib) - bal%sfc_pstart + 1 - start_pos + 1
      else
        num_elem = domain%blocks%length - start_pos + 1
      end if
      if(num_elem>0) then
        if(first_rank_in_cgpart) then
          ! first rank in a coarse partition additionally sends previous block
          start_pos = start_pos - 1
          num_elem = num_elem + 1
          first_rank_in_cgpart = .false.
        end if
#ifdef FD4_SFC_DISTR_VERBOSE
        write(sfcdbg+rank,*) 'send weights: group ',cgpart
        write(sfcdbg+rank,*) 'send weights: l pos ',start_pos, start_pos+num_elem-1
        write(sfcdbg+rank,*) 'send weights: g pos ',start_pos+bal%sfc_pstart-1, start_pos+bal%sfc_pstart-1+num_elem-1
#endif
        crank = cgpart*groupsz
        if(crank == rank) then
          ! do not send msg to myself, prepare local copy
          local_copy_start = start_pos
          local_copy_num = num_elem
        else
          ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
          if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
          tag = mintag + INT(num_elem)
          call MPI_Isend(bal%sfcd_blkweights_ps(start_pos), num_elem, MPI_R8K, crank, &
                          tag, mpi%comm, request, err)
          err = err - MPI_SUCCESS
          FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
          call MPI_Request_free(request, err)
          err = -1
          if(domain%stat_enabled) then
            call gettime_ofday(timing_us1)
            domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
          end if
        end if
      end if
      if(cgpart<cgpart1) start_pos = start_pos + num_elem
      if(start_pos>domain%blocks%length) exit
      first_rank_in_cgpart = .true.
      ib = ib + 1
    end do

#ifdef FD4_SFC_DISTR_VERBOSE
    if(rank==0) print *,partition
#endif

    ! allocate vector for all block weights of the local group
    if(.not. associated(bal%blkweights_ps)) then
      allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
      err = -1
    else
      if( ubound(bal%blkweights_ps,1) <      ncgblocks  .or. &
          ubound(bal%blkweights_ps,1) > 16 * ncgblocks ) then
        deallocate(bal%blkweights_ps)
        allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
        err = -1
      end if
    end if
    blkweights_ps => bal%blkweights_ps


    ! group masters receive block weights
    if( bal%sfcd_mpi%rank == 0 .and. ncgblocks>0 ) then

      ! reset the stack of requests
      call stack_create(mpi%requests, cgborders(4)-cgborders(3)+1)
      first_rank_in_cgpart = .true.
      start_pos = cgborders(1)
      crank = cgborders(3)
#ifdef FD4_SFC_DISTR_DEBUG
      ! find rank which owns block at start_pos
      call fd4_part_sfc_find_partition(partition, start_pos, crank)
      if(crank/=cgborders(3)) then
        write(sfcdbg+rank,*) 'EMERGENCY ABORT 2'
        write(sfcdbg+rank,*) ' crank: ',crank
        write(sfcdbg+rank,*) ' cgborders(3): ',cgborders(3)
        emergency_abort=1
      end if
#endif
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      ! loop over ranks to recv from
      do
        num_elem = MIN(partition(crank+1) - start_pos, cgborders(2) - start_pos + 1)
        if(num_elem>0) then
          if(first_rank_in_cgpart) then
            ! first rank in a coarse partition additionally sends previous block
            start_pos = start_pos - 1
            num_elem = num_elem + 1
#ifndef FD4_SFC_DISTR_DEBUG
            first_rank_in_cgpart = .false.
#endif
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'recv weights: rank ',crank
          write(sfcdbg+rank,*) 'recv weights: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
          write(sfcdbg+rank,*) 'recv weights: g pos ',start_pos, start_pos+num_elem-1
#endif
          if(crank == rank) then
            ! copy from own local weight vector
#ifdef FD4_SFC_DISTR_DEBUG
  if(num_elem/=local_copy_num) then
  !if(.true.) then
    write(sfcdbg+rank,*) 'EMERGENCY ABORT'
    write(sfcdbg+rank,*) 'sfcd group/rank: ',bal%sfcd_mygroup, bal%sfcd_mpi%rank
    write(sfcdbg+rank,*) 'partition:'
    do ib=0,nprocs
      write(sfcdbg+rank,*) ib,partition(ib)
    end do
    write(sfcdbg+rank,*) 'local block weights:'
    do ib=0,domain%blocks%length
      write(sfcdbg+rank,*) ib, bal%sfc_pstart+ib-1, bal%sfcd_blkweights_ps(ib)
    end do
    write(sfcdbg+rank,*) 'coarse scan ', cgpart0, cgpart1
    write(sfcdbg+rank,*) 'cgborders:',cgborders(1:2)
    do ib=1,cgborders_found%pos
      write(sfcdbg+rank,*) 'cgborders_found: ',ib,cgborders_found%a(ib)
    end do
    write(sfcdbg+rank,*) 'num_elem:',num_elem
    write(sfcdbg+rank,*) 'local_copy_num:',local_copy_num
    write(sfcdbg+rank,*) 'local_copy_start:',local_copy_start
    write(sfcdbg+rank,*) 'first_rank_in_cgpart:',first_rank_in_cgpart
    write(sfcdbg+rank,*) 'recv weights: rank ',crank
    write(sfcdbg+rank,*) 'recv weights: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
    write(sfcdbg+rank,*) 'recv weights: g pos ',start_pos, start_pos+num_elem-1
    !
    ib = 1
    start_pos = 1
    first_rank_in_cgpart = (partition(rank)==1 .or. exact_part_match)
    do cgpart = cgpart0-1, cgpart1
      if(cgpart<cgpart1) then
        num_elem = cgborders_found%a(ib) - bal%sfc_pstart + 1 - start_pos + 1
      else
        num_elem = domain%blocks%length - start_pos + 1
      end if
      if(num_elem>0) then
        if(first_rank_in_cgpart) then
          ! first rank in a coarse partition additionally sends previous block
          start_pos = start_pos - 1
          num_elem = num_elem + 1
        end if
        write(sfcdbg+rank,*) 'send weights: group ',cgpart
        write(sfcdbg+rank,*) 'send weights: rank  ',cgpart*groupsz
        write(sfcdbg+rank,*) 'send weights: first ',first_rank_in_cgpart, exact_part_match
        write(sfcdbg+rank,*) 'send weights: l pos ',start_pos, start_pos+num_elem-1
        write(sfcdbg+rank,*) 'send weights: g pos ',start_pos+bal%sfc_pstart-1, start_pos+bal%sfc_pstart-1+num_elem-1
        first_rank_in_cgpart = .false.
      end if
      if(cgpart<cgpart1) start_pos = start_pos + num_elem
      if(start_pos>domain%blocks%length) exit
      first_rank_in_cgpart = .true.
      ib = ib + 1
    end do
    FD4_L2_INFO('num_elem/=local_copy_num')
    emergency_abort=1
    exit
  end if
#else
            FD4_L1_ASSERT(num_elem/=local_copy_num,'num_elem/=local_copy_num')
#endif
            blkweights_ps(start_pos-cgborders(1)+1:start_pos-cgborders(1)+num_elem) = &
              bal%sfcd_blkweights_ps(local_copy_start:local_copy_start+num_elem-1)
          else
            ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
            tag = mintag + INT(num_elem)
            call MPI_Irecv(blkweights_ps(start_pos-cgborders(1)+1), num_elem, MPI_R8K, crank, &
                                         tag, mpi%comm, request, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
            err = -1
            call stack_push_fast(mpi%requests, request)
          end if
        end if
#ifdef FD4_SFC_DISTR_DEBUG
        first_rank_in_cgpart = .false.
#endif
        if(num_elem == cgborders(2) - start_pos + 1) exit
        start_pos = start_pos + num_elem
        crank = crank + 1
        FD4_L1_ASSERT(crank>=nprocs,'bug in recv loop')
      end do
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if

      if(crank/=cgborders(4)) then
#ifdef FD4_SFC_DISTR_DEBUG
        write(sfcdbg+rank,*) 'EMERGENCY ABORT 2'
        write(sfcdbg+rank,*) ' crank: ',crank
        write(sfcdbg+rank,*) ' cgborders(4): ',cgborders(4)
        emergency_abort=1
#else
        FD4_L1_ERROR('!rank2: '//i2s(crank-1)//i2s(cgborders(4)))
#endif
      end if

    end if ! group masters receive block weights

#ifdef FD4_SFC_DISTR_DEBUG
    call MPI_Allreduce(MPI_IN_PLACE, emergency_abort, 1, MPI_I_K, MPI_MAX, mpi%comm, err)
    if(emergency_abort>0) then
      err = -42
      return
    end if
#endif

    ! wait for recv to complete
    if(mpi%requests%pos>0) then
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if
    end if

    ! group master creates local prefix sum (i.e. substract first value from all)
    if( bal%sfcd_mpi%rank == 0 .and. ncgblocks>0 ) then
      blkweights_ps(1:ncgblocks) = blkweights_ps(1:ncgblocks) - blkweights_ps(0)
    end if
    blkweights_ps(0) = 0.0_r8k

    ! group master bcasts weights to all group members
    if( ncgblocks>0 ) then
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      call MPI_Bcast(blkweights_ps(1), ncgblocks, MPI_R8K, 0, bal%sfcd_mpi%comm, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Bcast failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if
    end if

    err = 0

  end subroutine fd4_part_sfcd_exch_weights


  !! Exchange block weights. Less p2p traffic than $fd4_part_sfcd_exch_weights.
  !! This is $FD4_PART_SFCD_WGTX_COLL.
  !!
  !! Method:
  !! - weights from non-group members are sent to first/last group member with p2p messages
  !! - weights are exchanged within group with MPI_Allgatherv (in-place) or MPI_Gatherv (if $toall is false)
  !!
  subroutine fd4_part_sfcd_exch_weights_coll(domain, cgpart0, cgborders_found, cgborders, exact_part_match, toall, err)
    type(fd4_domain), intent(inout) :: domain              !! the domain
    integer(i_k)    , intent(in)    :: cgpart0             !! lower bound for search of local coarse border matches
    type(stack)     , intent(in)    :: cgborders_found     !! stack containing local matches of coarse partition borders
    integer(i_k)    , intent(in)    :: cgborders(4)        !! partitions borders/ranks of own coarse partition
    logical         , intent(in)    :: exact_part_match    !! first local block exactly starts new coarse partition?
    logical         , intent(in)    :: toall               !! gather weights to all (or only masters)?
    integer(i_k)    , intent(out)   :: err                 !! error status: 0...ok
    target :: domain
    
    integer :: rank, nprocs
    integer(i_k) :: ib, crank, grank
    real(r8k), pointer :: blkweights_ps(:)     ! prefix sum of sfc sorted block weights
    integer(i_k), pointer :: partition(:)      ! partition vector (start indices)
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: timing_us0mpi, timing_us1
    type(fd4_domain_balance), pointer :: bal
    integer(i_k) :: cgpart1, cgpart
    integer :: request, tag
    integer(i_k) :: ncgblocks, start_pos, num_elem, groupsz, rankloop(2)
    integer(i_k) :: local_copy_start, local_copy_num
    ! is the first block to send the first block in a coarse partition?
    logical :: first_rank_in_cgpart
    ! add this to message tag so that overlap with lower tags not possible
    integer, parameter :: mintag = 20
    logical :: cgborders_outside_group

    err = -1

    mpi    => domain%mpi
    bal    => domain%balance
    rank    = mpi%rank
    nprocs  = mpi%nprocs
    partition => domain%balance%partition
    groupsz = bal%sfcd_groupsz
    cgpart1 = cgpart0 + cgborders_found%pos - 1
    ncgblocks = cgborders(2) - cgborders(1) + 1
    mpi%requests%pos = 0

    call fd4_mpi_realloc_array(mpi%recvcnt, groupsz, 0, err)
    FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')
    call fd4_mpi_realloc_array(mpi%recvdis, groupsz, 0, err)
    FD4_L1_ASSERT(err/=0,'fd4_mpi_realloc_array failed')

    local_copy_start = 0
    local_copy_num = 0

    ! send block weights to first/last group members of other groups
    if(domain%blocks%length>0) then
    
      ib = 1
      start_pos = 1
      ! first_rank_in_cgpart is true before the loop if
      !   this ranks current partition starts with first block of any coarse partition (not only own coarse partition!!)
      first_rank_in_cgpart = (partition(rank)==1 .or. exact_part_match)  ! does not work with H2 coarse search
      !first_rank_in_cgpart = (bal%sfc_pstart==cgborders(1))
#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'C send weights: pstart, cgborders(1) ',bal%sfc_pstart, cgborders(1)
      write(sfcdbg+rank,*) 'C send weights: first_rank_in_cgpart ',first_rank_in_cgpart
      write(sfcdbg+rank,*) 'C send weights: partition(rank)==1   ',partition(rank)==1
      write(sfcdbg+rank,*) 'C send weights: exact_part_match     ',exact_part_match
#endif

      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

      do cgpart = cgpart0-1, cgpart1
        if(cgpart<cgpart1) then
          num_elem = cgborders_found%a(ib) - bal%sfc_pstart + 1 - start_pos + 1
        else
          num_elem = domain%blocks%length - start_pos + 1
        end if
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'C send weights? group ',cgpart, num_elem
#endif
        if(num_elem>0) then
          if(first_rank_in_cgpart) then
            ! first rank in a coarse partition additionally sends previous block
            start_pos = start_pos - 1
            num_elem = num_elem + 1
            first_rank_in_cgpart = .false.
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'C send weights: group ',cgpart
          write(sfcdbg+rank,*) 'C send weights: l pos ',start_pos, start_pos+num_elem-1
          write(sfcdbg+rank,*) 'C send weights: g pos ',start_pos+bal%sfc_pstart-1, start_pos+bal%sfc_pstart-1+num_elem-1
#endif
          if(cgpart==bal%sfcd_mygroup) then
            ! this part of my blocks stays within my group: prepare local copy for allgatherv
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'C send weights: prepare local copy', start_pos, num_elem
#endif
            ! prepare local copy
            local_copy_start = start_pos
            local_copy_num = num_elem
          else
            crank = cgpart*groupsz
            ! send to last group member my group comes after group to send to
            if(cgpart<bal%sfcd_mygroup) crank = crank + groupsz-1
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'C send weights:  p2p to rank ', crank, '   size ',num_elem
#endif
            ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
            tag = mintag + INT(num_elem)
            call MPI_Isend(bal%sfcd_blkweights_ps(start_pos), num_elem, MPI_R8K, crank, &
                            tag, mpi%comm, request, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
            call MPI_Request_free(request, err)
            err = -1
          end if
        end if ! num_elem>0
        if(cgpart<cgpart1) start_pos = start_pos + num_elem
        if(start_pos>domain%blocks%length) exit
        first_rank_in_cgpart = .true.
        ib = ib + 1
      end do

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if

    end if ! domain%blocks%length>0

#ifdef FD4_SFC_DISTR_VERBOSE
    if(rank==0) print *,partition
#endif

    ! allocate vector for all block weights of the local group
    if(.not. associated(bal%blkweights_ps)) then
      allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
      FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
      err = -1
    else
      if( ubound(bal%blkweights_ps,1) <      ncgblocks  .or. &
          ubound(bal%blkweights_ps,1) > 16 * ncgblocks ) then
        deallocate(bal%blkweights_ps)
        allocate(bal%blkweights_ps(0:MAX(2*ncgblocks,16)), stat=err)
        FD4_L1_ASSERT(err/=0,'allocate bal%blkweights_ps')
        err = -1
      end if
    end if
    blkweights_ps => bal%blkweights_ps

    ! first and last group members receive block weights
    if( (bal%sfcd_mpi%rank == 0 .or. bal%sfcd_mpi%rank == groupsz-1) .and. ncgblocks>0 ) then

      if( bal%sfcd_mpi%rank == 0 ) then
        ! first group member receives from ranks before group
        rankloop(1) = cgborders(3)                         ! first rank to recv from
        rankloop(2) = MIN(cgborders(4), rank-1)            ! last  rank to recv from
        start_pos = cgborders(1)                           ! first block to recv
      else
        ! last group member receives from ranks after group
        rankloop(1) = MAX(cgborders(3), rank+1)            ! first rank to recv from
        rankloop(2) = cgborders(4)                         ! last  rank to recv from
        start_pos = MAX(cgborders(1), partition(rank+1))   ! first block to recv
      end if

#ifdef FD4_SFC_DISTR_VERBOSE
      write(sfcdbg+rank,*) 'C recv weights: rankloop ',rankloop(1:2)
      write(sfcdbg+rank,*) 'C recv weights: start_pos ',start_pos
#endif

      if(rankloop(2) >= rankloop(1)) then

        ! reset the stack of requests
        call stack_create(mpi%requests, rankloop(2)-rankloop(1)+1)
        first_rank_in_cgpart = ( start_pos == cgborders(1) )

        if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

        ! loop over ranks to recv from
        do crank = rankloop(1), rankloop(2)
          num_elem = MIN(partition(crank+1) - start_pos, cgborders(2) - start_pos + 1)
          if(num_elem>0) then
            if(first_rank_in_cgpart) then
              ! first rank in a coarse partition additionally sends previous block
              start_pos = start_pos - 1
              num_elem = num_elem + 1
              first_rank_in_cgpart = .false.
            end if
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'C recv weights: rank ',crank, '   size ',num_elem
            write(sfcdbg+rank,*) 'C recv weights: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
            write(sfcdbg+rank,*) 'C recv weights: g pos ',start_pos, start_pos+num_elem-1
#endif
            ! go! use msg size as tag to force deadlocks if msg size differs at sender and receiver
            tag = mintag + INT(num_elem)
            call MPI_Irecv(blkweights_ps(start_pos-cgborders(1)+1), num_elem, MPI_R8K, crank, &
                                         tag, mpi%comm, request, err)
            err = err - MPI_SUCCESS
            FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
            err = -1
            call stack_push_fast(mpi%requests, request)
          end if
          start_pos = start_pos + num_elem
        end do ! rankloop

        if(domain%stat_enabled) then
          call gettime_ofday(timing_us1)
          domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
        end if

      end if ! rankloop(2) >= rankloop(1)

    end if ! first and last group members receive block weights


#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
    flush(sfcdbg+rank)
    ferr=fsync(fnum(sfcdbg+rank))
#endif

    ! wait for recv to complete
    if(mpi%requests%pos>0) then
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
      err = -1
      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if
    end if

    ! distribute block weights within group to all ranks with MPI_Allgatherv
    if( ncgblocks>0 ) then
    
      ! loop over ranks of group and prepare MPI_Allgatherv
      mpi%recvdis(0) = 0
      crank = bal%sfcd_mygroup*groupsz ! global rank
      do grank = 0, groupsz-1
        cgborders_outside_group = &
              (grank==0         .and. crank>cgborders(4)) .or. &  ! upper border before 1st group rank
              (grank==groupsz-1 .and. crank<cgborders(3))         ! lower border after last group rank
#ifdef FD4_SFC_DISTR_VERBOSE
        write(sfcdbg+rank,*) 'C allgatherv: rank ',crank, grank
        write(sfcdbg+rank,*) 'C allgatherv: cgborders_outside_group ',cgborders_outside_group
#endif
        if( (crank>=cgborders(3) .and. crank<=cgborders(4)) .or. cgborders_outside_group ) then
          ! this rank owns block weights which are part of new coarse partition
          ! first group rank owns all block weights before group
          if(grank==0) then
            start_pos = cgborders(1)
          else
            start_pos = MAX(partition(crank), cgborders(1))
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'C allgatherv: first_rank_in_cgpart ',start_pos == cgborders(1)
#endif
          ! first rank in a coarse partition additionally sends previous block
          if( start_pos == cgborders(1) ) start_pos = start_pos - 1
          ! last group rank owns all block weights after group
          if(grank==groupsz-1) then
            num_elem = cgborders(2) - start_pos + 1
          else
            num_elem = MIN(partition(crank+1)-1, cgborders(2)) - start_pos + 1
          end if
#ifdef FD4_SFC_DISTR_VERBOSE
          if(num_elem<=0) write(sfcdbg+rank,*) 'C allgatherv: num_elem<=0 ',num_elem
#endif
          num_elem = MAX(0, num_elem)
          mpi%recvcnt(grank) = num_elem
#ifdef FD4_SFC_DISTR_VERBOSE
          write(sfcdbg+rank,*) 'C allgatherv: l pos ',start_pos-cgborders(1)+1, start_pos-cgborders(1)+1+num_elem-1
          write(sfcdbg+rank,*) 'C allgatherv: g pos ',start_pos, start_pos+num_elem-1
#endif
          if(crank == rank .and. .not. cgborders_outside_group) then
            ! local copy
            ! only consider weights of own blocks, not those that were received with p2p msg from other group ranks
            start_pos = MAX(partition(crank), cgborders(1))
            if( start_pos == cgborders(1) ) start_pos = start_pos - 1
            num_elem = MIN(partition(crank+1)-1, cgborders(2)) - start_pos + 1
#ifdef FD4_SFC_DISTR_VERBOSE
            if(num_elem<=0) write(sfcdbg+rank,*) 'C allgatherv: local copy num_elem<=0 ',num_elem
#endif
            num_elem = MAX(0, num_elem)
#ifdef FD4_SFC_DISTR_VERBOSE
            write(sfcdbg+rank,*) 'C allgatherv: local copy ',num_elem, local_copy_num, local_copy_start
#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
            flush(sfcdbg+rank)
            ferr=fsync(fnum(sfcdbg+rank))
#endif
#endif
            !FD4_L2_INFO(i2s(num_elem)//i2s(local_copy_num))
            FD4_L1_ASSERT(num_elem/=local_copy_num,'num_elem/=local_copy_num')
            if(num_elem>0) then
              blkweights_ps(start_pos-cgborders(1)+1:start_pos-cgborders(1)+num_elem) = &
                bal%sfcd_blkweights_ps(local_copy_start:local_copy_start+num_elem-1)
            end if
          end if
        else
          mpi%recvcnt(grank) = 0
        end if
        if(grank>0) mpi%recvdis(grank) = mpi%recvdis(grank-1) + mpi%recvcnt(grank-1)
#ifdef FD4_SFC_DISTR_VERBOSE
        write(sfcdbg+rank,*) 'C allgatherv: recvcnt ',mpi%recvcnt(grank)
        write(sfcdbg+rank,*) 'C allgatherv: recvdis ',mpi%recvdis(grank)
#endif
        crank = crank + 1
      end do

#ifdef FD4_SFC_DISTR_VERBOSE_SYNC
      flush(sfcdbg+rank)
      ferr=fsync(fnum(sfcdbg+rank))
#endif

      if(mpi%recvcnt(groupsz-1)+mpi%recvdis(groupsz-1)/=ncgblocks+1) then
        FD4_L1_ERROR('recvcnt / recvdis error '//i2s(ncgblocks))
      end if
      
#ifdef FD4_SFC_DISTR_DEBUG
      !!DEBUG: check if all ranks in a group have same recvcnt for MPI_Allgatherv
      call mpi_allreduce(mpi%recvcnt(0),mpi%recvdis(0),groupsz,MPI_I_K,MPI_MAX,bal%sfcd_mpi%comm,err)
      if(ANY(mpi%recvcnt(0:groupsz-1)/=mpi%recvdis(0:groupsz-1))) then
        print *,rank,mpi%recvcnt(0:groupsz-1)
        print *,rank,mpi%recvdis(0:groupsz-1)
        FD4_L1_ERROR('recvcnt error')
      end if
      ! restore recvdis
      mpi%recvdis(0) = 0
      do grank = 1, groupsz-1
        mpi%recvdis(grank) = mpi%recvdis(grank-1) + mpi%recvcnt(grank-1)
      end do
#endif


      ! finally exchange the block weights within the group
      
      if(domain%stat_enabled) call gettime_ofday(timing_us0mpi)

      if(toall) then
#ifdef FD4_NO_MPI_IN_PLACE
#warning compiling with FD4_NO_MPI_IN_PLACE
        ! allocate array for sending the block weights
        if( size(mpi%sendreal,1) < mpi%recvcnt(bal%sfcd_mpi%rank) ) then
          deallocate(mpi%sendreal)
          allocate(mpi%sendreal(2*mpi%recvcnt(bal%sfcd_mpi%rank)))
        end if
        mpi%sendreal(1:mpi%recvcnt(bal%sfcd_mpi%rank)) = &
          blkweights_ps(mpi%recvdis(bal%sfcd_mpi%rank):mpi%recvdis(bal%sfcd_mpi%rank)+mpi%recvcnt(bal%sfcd_mpi%rank)-1)
        call MPI_Allgatherv(mpi%sendreal(1), mpi%recvcnt(bal%sfcd_mpi%rank), MPI_R8K, &
                         blkweights_ps(0), mpi%recvcnt(0),  mpi%recvdis(0), MPI_R8K, bal%sfcd_mpi%comm, err )
#else
        call MPI_Allgatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, blkweights_ps(0), &
                            mpi%recvcnt(0),  mpi%recvdis(0), MPI_R8K, bal%sfcd_mpi%comm, err )
#endif
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')
      else
        mpi%recvcnt(0) = 0
        call MPI_Gatherv(blkweights_ps(mpi%recvdis(bal%sfcd_mpi%rank)), mpi%recvcnt(bal%sfcd_mpi%rank), MPI_R8K, &
                         blkweights_ps(0), mpi%recvcnt(0),  mpi%recvdis(0), MPI_R8K, 0, bal%sfcd_mpi%comm, err )
!        if(bal%sfcd_mpi%rank==0) then
!          call MPI_Gatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL, &
!                           blkweights_ps(0), mpi%recvcnt(0),  mpi%recvdis(0), MPI_R8K, 0, bal%sfcd_mpi%comm, err )
!        else
!          call MPI_Gatherv(blkweights_ps(mpi%recvdis(bal%sfcd_mpi%rank)), mpi%recvcnt(bal%sfcd_mpi%rank), MPI_R8K, &
!                           blkweights_ps(0), mpi%recvcnt(0),  mpi%recvdis(0), MPI_R8K, 0, bal%sfcd_mpi%comm, err )
!        end if
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Gatherv failed')
      end if

      if(domain%stat_enabled) then
        call gettime_ofday(timing_us1)
        domain%stat%values(bal%stat_mpi_us) = domain%stat%values(bal%stat_mpi_us) + (timing_us1-timing_us0mpi)
      end if


      ! create local prefix sum (i.e. substract first value from all)
      if(toall .or. bal%sfcd_mpi%rank==0) then
        blkweights_ps(1:ncgblocks) = blkweights_ps(1:ncgblocks) - blkweights_ps(0)
        blkweights_ps(0) = 0.0_r8k
      end if
      

    end if ! ncgblocks>0
    

    err = 0
       
  end subroutine fd4_part_sfcd_exch_weights_coll


#endif

end module fd4_part_sfcd_mod
