! 2013-02-22
! Program to test serial and parallel FD4 partitioning routines (SFC)
! Uses original FD4
!
! compile with:
!  FD4FLAGS += -DFD4_SFC_DISTR
!  FD4FLAGS += -DFD4_REAL8_WEIGHTS
!  FD4FLAGS += -DFD4_PART_1D_SCAN_HAN
! 
! Methods:
!   1 - 'simple1 '   H1 sequential
!   2 - 'simple2 '   H2 sequential
!   3 - 'binsrch '   r-BISECT sequential,  aka QBS
!   4 - 'parallel'   r-PAR-BISECT,         aka QBS*
!   5 - 'recbisct'   recursive bisection sequential, not optimized
!   6 - 'ebspinar'   EXACT-BISECT sequential (Pinar & Aykanat)
!   7 - 'h-distr1'   H1 parallel
!   8 - 'h-distr2'   H2 parallel
!   9 - 'rp-dst1 '   r-DISTR-BISECT, parallel, H1 coarse, r-PAR-BISECT fine, dspv=allgather
!  10 - 'rp-dst2 '   r-DISTR-BISECT, parallel, H2 coarse, r-PAR-BISECT fine, dspv=allgather
!  11 - 'rs-dst1 '   r-DISTR-BISECT, parallel, H1 coarse, r-BISECT fine    , dspv=scatter+allgather
!  12 - 'rs-dst2 '   r-DISTR-BISECT, parallel, H2 coarse, r-BISECT fine    , dspv=scatter+allgather
!  13 - 'rp+dst1 '   r-DISTR-BISECT, parallel, H1 coarse, r-PAR-BISECT fine, dspv=allgather+bcast
!  14 - 'rp+dst2 '   r-DISTR-BISECT, parallel, H2 coarse, r-PAR-BISECT fine, dspv=allgather+bcast,    aka HIER*
!  15 - 'rs+dst1 '   r-DISTR-BISECT, parallel, H1 coarse, r-BISECT fine,     dspv=allgather+bcast
!  16 - 'rs+dst2 '   r-DISTR-BISECT, parallel, H2 coarse, r-BISECT fine,     dspv=allgather+bcast,    aka HIER
!  17 - 'h1+dst1 '   r-DISTR-BISECT, parallel, H1 coarse, H1 fine,           dspv=allgather+bcast
!  18 - 'h2+dst2 '   r-DISTR-BISECT, parallel, H2 coarse, H2 fine,           dspv=allgather+bcast
!  19 - 'binsrchW'   r-BISECT sequential, maxweight is computed at root
!  20 - 'paralleW'   r-PAR-BISECT, maxweight is computed at root
!  21 - 'ebspinaW'   EXACT-BISECT sequential (Pinar & Aykanat), maxweight is computed at root
!  22 - 'ebs+dst2'   r-DISTR-BISECT, parallel, H2 coarse, EXACT-BISECT fine, dspv=allgather+bcast

!#define UNFORMATTED_INPUT

#if (defined(VTRACE))
# define VT_USER_START(n) call VT_User_start__((n), __FILE__, __LINE__)
# define VT_USER_END(n) call VT_User_end__((n))
#else
# define VT_USER_START(n)
# define VT_USER_END(n)
#endif

program partbench

  use fd4_mod
  use fd4_stat_mod
  use fd4_mpi_mod
  use fd4_domain_mod, only: fd4_domain_balance
  use timing_mod
  
  implicit none
  
  include 'mpif.h'
  
  integer(i_k) :: blnum(3), err, dbnd(3,2), dumpcounter, nblocks, ncycles, crank, wrank
  integer(i_k) :: blnum_file(3), multi(2), sfctype, method, step_select(3)
  integer(i_k) :: sfcd_groupsz, p1d_root
  real(r8k), allocatable :: blkweights3d(:,:,:), localwgts(:)
  real(r8k) :: maxweight, maxweight_loc, weight_balance, balance, sumload, maxload, rndm, rndm_perturbation, avgbalance
  integer :: tmp, min_blocks, min_blocks_rank
  integer(i8k) :: tloc, tsum
  integer :: req_steps, i, j, k, nlocblocks
  character(8), parameter :: method_str(1:22)=(/'simple1 ', 'simple2 ', 'binsrch ', 'parallel', &
                                                'recbisct', 'ebspinar', 'h-distr1', 'h-distr2', &
                                                'rp-dst1 ', 'rp-dst2 ', 'rs-dst1 ', 'rs-dst2 ', &
                                                'rp+dst1 ', 'rp+dst2 ', 'rs+dst1 ', 'rs+dst2 ', &
                                                'h1+dst1 ', 'h2+dst2 ', &
                                                'binsrchW', 'paralleW', 'ebspinaW', &
                                                'ebs+dst2'/)
  integer(i_k), parameter :: method_ids(1:22)=(/FD4_BALANCE_HSFC2_SIMPLE,  FD4_BALANCE_HSFC2_SIMPLE2, &
                                                FD4_BALANCE_HSFC2_BINSRCH, FD4_BALANCE_HSFC2_PARALLEL, &
                                                FD4_BALANCE_HSFC2_RB, FD4_BALANCE_HSFC2_EBS, &
                                                FD4_BALANCE_HSFC_DSIMPLE, FD4_BALANCE_HSFC_DSIMPLE2, &
                                                FD4_BALANCE_HSFC_DISTR, FD4_BALANCE_HSFC_DISTR2, &
                                                FD4_BALANCE_HSFC_DISTR, FD4_BALANCE_HSFC_DISTR2, &
                                                FD4_BALANCE_HSFC_DISTR, FD4_BALANCE_HSFC_DISTR2, &
                                                FD4_BALANCE_HSFC_DISTR, FD4_BALANCE_HSFC_DISTR2, &
                                                FD4_BALANCE_HSFC_DISTR, FD4_BALANCE_HSFC_DISTR2, &
                                                FD4_BALANCE_HSFC2_BINSRCH, FD4_BALANCE_HSFC2_PARALLEL, &
                                                FD4_BALANCE_HSFC2_EBS, &
                                                FD4_BALANCE_HSFC_DISTR2 /)
  logical,      parameter :: method_ser(1:22)=(/.true. , .true. , .true. , .false., &
                                                .true. , .true. , .false., .false., &
                                                .false., .false., .false., .false., &
                                                .false., .false., .false., .false., &
                                                .false., .false., &
                                                .true. , .false., .true., &
                                                .false. /)
  character(8), parameter :: sfc_str(1:2)=(/'hilbert ','morton  '/)
  integer(i_k), parameter :: sfc_ids(1:2)=(/FD4_PART_SFC_HILBERT,FD4_PART_SFC_MORTON/)
  real(r8k) :: avgweight
  
  real(r8k) :: lb_target
  integer :: steps_sum
  integer(i8k) :: migrated_blocks, partition_shifts ! migration statistics

  ! domain stuff
  type(fd4_domain), target :: domain
  type(fd4_vartab) :: vartab(1)
  integer :: dsize(3,2), nghosts(3)
  logical :: periodic(3)
  type(fd4_mpi), pointer :: mpi => NULL()
  type(fd4_domain_balance), pointer :: bal => NULL()
  integer(i_k), pointer :: partition(:) => NULL(), partition_old(:) => NULL()
  ! iter stuff
  type(fd4_iter) :: iter
  integer(i_k) :: offset(3)
  ! stats stuff
  type(fd4_balance_statistics) :: bal_stats
  integer(i8k), allocatable :: stat_values(:,:), stat_values_coll(:,:,:)
  real(r8k) :: heap_usage
  ! i/o groups
  integer :: io_group_size, io_group, io_comm, io_rank
  ! config bcast
  integer(i8k) :: config_i(11)
  real(r8k) :: config_r(2)
  integer :: out

  mpi => domain%mpi
  bal => domain%balance

  call MPI_Init(err)
  call MPI_Comm_rank(MPI_COMM_WORLD,mpi%rank,err)
  call MPI_Comm_size(MPI_COMM_WORLD,mpi%nprocs,err)
  
  !call MPI_Barrier(MPI_COMM_WORLD, err)
  !if(mpi%rank==0) then
  !  write(50,'(A8)') 'started'
  !  close(50)
  !end if
  
  if(mpi%rank==0) then
    ! read config file
    read(40,'(2(I5))') config_i(1:2)       ! multi(1:2)
    read(40,'(F5.3)')  config_r(1)         ! rndm_perturbation
    read(40,'(I2)')    config_i(3)         ! sfctype
    read(40,'(I2)')    config_i(4)         ! method
    read(40,'(I7)')    tmp                 ! nprocs (ignored)
    read(40,'(F6.4)')  config_r(2)         ! lb_target
    read(40,'(3(I5))') config_i(5:7)       ! step_select(1:3)  ! start, measurement start (first runs to warm up), stop
    read(40,'(I7)')    config_i(8)         ! sfcd_groupsz
    read(40,'(I7)')    config_i(9)         ! p1d_root
    read(40,'(I7)', iostat=tmp)  config_i(10) ! io_group_size
    if(tmp/=0) config_i(10) = 9999999
    read(40,'(I7)', iostat=tmp)  config_i(11) ! out
    if(tmp/=0) config_i(11) = 6  ! stdout
    close(40)
  end if
  
  ! bcast config
  call MPI_Bcast(config_i(1), 11, MPI_I8K, 0, MPI_COMM_WORLD, err)
  call MPI_Bcast(config_r(1),  2, MPI_R8K, 0, MPI_COMM_WORLD, err)
  ! set config variables
  multi(1:2) = config_i(1:2)
  rndm_perturbation = config_r(1)
  sfctype = config_i(3)
  method = config_i(4)
  lb_target = config_r(2)
  step_select(1:3) = config_i(5:7)
  sfcd_groupsz = config_i(8)
  p1d_root = config_i(9)
  io_group_size = config_i(10)
  out = config_i(11)
  
  
  ! create i/o groups of size io_group_size
  if(io_group_size>=mpi%nprocs) then
    io_group = 0
    io_comm = MPI_COMM_WORLD
    io_rank = mpi%rank
  elseif(io_group_size>1) then
    io_group = mpi%rank / io_group_size
    call MPI_Comm_split(MPI_COMM_WORLD, io_group, mpi%rank, io_comm, err)
    call MPI_Comm_rank(io_comm,io_rank,err)
  else
    io_group = mpi%rank
    io_comm = MPI_COMM_NULL
    io_rank = 0
  end if

  ncycles=0
  avgbalance=0.0_r8k
  if(sfctype<1 .or. sfctype>2) sfctype=1
  if(method<1 .or. method>22) method=1
  sfcd_groupsz = MIN(sfcd_groupsz, mpi%nprocs/2)
  sfcd_groupsz = MAX(sfcd_groupsz, 2)
  p1d_root = MIN(0, p1d_root)              ! p1d_root must be 0 or disabled (<0)
  if(method== 4 .or. method==20 .or. method== 7 .or. method== 8) p1d_root=-1 ! disable serial computation for parallel methods
  if(method==13 .or. method==14 .or. method== 9 .or. method==10) p1d_root=-1 ! disable serial computation for parallel methods
 
  migrated_blocks = 0_i8k
  partition_shifts = 0_i8k
  steps_sum = 0

  call init_random_seed(726183+mpi%rank)

  do dumpcounter = step_select(1),step_select(3)
  
    !! very first iteration?
    if(dumpcounter==step_select(1)) then
      if(io_rank==0) then
#ifdef UNFORMATTED_INPUT
        read(2000+dumpcounter) blnum_file(1:3), dbnd(1:3,1:2)
#else
        read(1000+dumpcounter,'(3(I7))') blnum_file(1:3)
        read(1000+dumpcounter,'(6(I7))') dbnd(1:3,1:2)
#endif
      end if
      if(io_group_size>1) then
        call MPI_Bcast(blnum_file, 3, MPI_I_K, 0, io_comm, err)
      end if
    else
      if(io_rank==0) then
        ! ignore blnum_file for later iterations
#ifdef UNFORMATTED_INPUT
        read(2000+dumpcounter) dbnd(1:3,1), dbnd(1:3,1:2)
#else
        read(1000+dumpcounter,'(3(I7))') dbnd(1:3,1)
        read(1000+dumpcounter,'(6(I7))') dbnd(1:3,1:2)
#endif
      end if
    end if
    
    blnum = blnum_file
    blnum(1:2) = blnum(1:2) * multi(1:2)  ! replicate domain in x and y by multi(1:2)
    nblocks = PRODUCT(blnum)
  
    !! very first iteration?
    if(dumpcounter==step_select(1)) then

      !! setup FD4 domain  

      vartab(1) = fd4_vartab('a',  1,  1, .false., 0.0_r8k, FD4_NOTHRES, FD4_CELLC )
      dsize(1:3,1) = 1
      dsize(1:3,2) = blnum
      nghosts(1:3) = 0
      periodic(1:3) = .true.

      call fd4_domain_create(domain, blnum, dsize, vartab, nghosts, periodic, MPI_COMM_WORLD, err)
      if(err/=0) then
        write(out,*) 'fd4_domain_create failed'
        call MPI_Abort(MPI_COMM_WORLD, 1, err)
      end if

      call fd4_balance_params(domain, method = method_ids(method) )
      ! opt_lbtol>1.0 forces load balancing even if balance is 1.0
      call fd4_balance_params(domain, opt_lbtol = 1.1_r4k)
      call fd4_balance_params(domain, opt_accuracy = lb_target)
      call fd4_balance_params(domain, opt_blkmap_update = .false.)
      call fd4_balance_params(domain, opt_auto = .false.)
      call fd4_balance_params(domain, opt_sfctype = sfc_ids(sfctype) )
      call fd4_balance_params(domain, opt_binsteps = 42 )
      call fd4_balance_params(domain, opt_par_pfxs = .true. )
      call fd4_balance_params(domain, opt_groupsz = sfcd_groupsz )
      call fd4_balance_params(domain, opt_p1d_root = p1d_root )
      if(method== 9 .or. method==10) then
        ! rp-dst1, rp-dst2
        call fd4_balance_params(domain, opt_dp1dmeth = FD4_PART_SFC_PARALLEL )
        call fd4_balance_params(domain, opt_dspvmeth = FD4_PART_SFCD_SCATTER_ALLGATHER )
      end if
      if(method==11 .or. method==12) then
        ! rs-dst1, rs-dst2
        call fd4_balance_params(domain, opt_dp1dmeth = FD4_PART_SFC_BINSRCH )
        call fd4_balance_params(domain, opt_dspvmeth = FD4_PART_SFCD_SCATTER_ALLGATHER )
      end if
      if(method==13 .or. method==14) then
        ! rp+dst1, rp+dst2
        call fd4_balance_params(domain, opt_dp1dmeth = FD4_PART_SFC_PARALLEL )
        call fd4_balance_params(domain, opt_dspvmeth = FD4_PART_SFCD_ALLGATHER_BCAST )
      end if
      if(method==15 .or. method==16) then
        ! rs+dst1, rs+dst2
        call fd4_balance_params(domain, opt_dp1dmeth = FD4_PART_SFC_BINSRCH )
        call fd4_balance_params(domain, opt_dspvmeth = FD4_PART_SFCD_ALLGATHER_BCAST )
      end if
      if(method==17) then
        ! h1+dst1
        call fd4_balance_params(domain, opt_dp1dmeth = FD4_PART_SFC_SIMPLE )
        call fd4_balance_params(domain, opt_dspvmeth = FD4_PART_SFCD_ALLGATHER_BCAST )
      end if
      if(method==18) then
        ! h2+dst2
        call fd4_balance_params(domain, opt_dp1dmeth = FD4_PART_SFC_SIMPLE2 )
        call fd4_balance_params(domain, opt_dspvmeth = FD4_PART_SFCD_ALLGATHER_BCAST )
      end if
      if(method==19 .or. method==20 .or. method==21) then
        ! do not communicate to get maxweight
        call fd4_balance_params(domain, opt_mwgtcomm = .false. )
      end if
      if(method==22) then
        ! ebs+dst2
        call fd4_balance_params(domain, opt_dp1dmeth = FD4_PART_SFC_EBS )
        call fd4_balance_params(domain, opt_dspvmeth = FD4_PART_SFCD_ALLGATHER_BCAST )
      end if

      call fd4_util_allocate_all_blocks(domain, err)
      if(err/=0) then
        write(out,*) 'fd4_util_allocate_all_blocks failed'
        call MPI_Abort(MPI_COMM_WORLD, 1, err)
      end if
      
      !! allocate array for all block weights from file
      allocate(blkweights3d(blnum_file(1),blnum_file(2),blnum_file(3)))
      
      !! allocate stats array (for each step)
      allocate(stat_values(size(domain%stat%values),step_select(2):step_select(3)))
      allocate(stat_values_coll(size(domain%stat%values),1,1))

      if(mpi%rank==0) write(out,'(A3,A10,A10,A5,A7,A3,A5,A8,A11,A11,A11,A10,A10,A12,A13,A4)') &
         'cnt','balance','maxload','minb','at','st','wbal','migblk','tsum/us', &
         't_pfxs', 't_exwgt', 't_p1dc', 't_bcbr', 't_p1d', 't_dspv', 'mem'

    end if
    
    nlocblocks = domain%blocks%length
        
    blkweights3d = 0.0_r8k
    
#ifdef UNFORMATTED_INPUT
    if(io_rank==0) then
      VT_USER_START('PARTBENCH::READ_WEIGHTS')
      read(2000+dumpcounter) blkweights3d
      close(2000+dumpcounter)
      VT_USER_END('PARTBENCH::READ_WEIGHTS')
    end if
#else
    if(io_rank==0) then
      VT_USER_START('PARTBENCH::READ_WEIGHTS')
      ! read weights
      do k=1,blnum_file(3)
        do j=1,blnum_file(2)
          do i=1,blnum_file(1)
            read(1000+dumpcounter,'(E15.8)') blkweights3d(i,j,k)
            ! reduce precision to force exact part matches in sfcd method
            !blkweights3d(i,j,k) = MAX(1.0, REAL(INT(blkweights3d(i,j,k)) / 10000, r8k) * 10000.0)
          end do
        end do
      end do
      close(1000+dumpcounter)
      VT_USER_END('PARTBENCH::READ_WEIGHTS')
    end if
#endif

    if(io_group_size>1) then
      call MPI_Bcast(blkweights3d, PRODUCT(blnum_file), MPI_R8K, 0, io_comm, err)
    end if
    
    ! set weight of all local blocks
    maxweight_loc = 0.0_r8k
    call fd4_iter_init(domain, iter)
    do while(associated(iter%cur))  
      call fd4_iter_offset(iter, offset) ! get offset from domain indexes to block-local indexes
      offset = offset + 1
      iter%cur%weight = REAL( get_replicate_weight(blkweights3d, blnum_file, offset(1), offset(2), offset(3)) , r4k)
      !iter%cur%weight = 1.0
      maxweight_loc = MAX(maxweight_loc, iter%cur%weight)
      call fd4_iter_next(iter)
    end do

    ! random perturbation of weight data
    if(rndm_perturbation>tiny(rndm_perturbation)) then
      maxweight_loc = 0.0_r8k
      call fd4_iter_init(domain, iter)
      do while(associated(iter%cur))  
        call random_number(rndm)
        iter%cur%weight = iter%cur%weight * (1.0_r8k+(rndm-0.5_r8k)*rndm_perturbation)
        maxweight_loc = MAX(maxweight_loc, iter%cur%weight)
        call fd4_iter_next(iter)
      end do
    end if
    
    
    ! reset statistics
    call fd4_stat_reset(domain%stat)

    !!
    !! Start 1D partitioning
    !!
    !! Input:  - distributed weight vector (different sizes and values in each rank)
    !!              from the iter%cur%weight's within the blocks fd4_part_sfc_weight2 sets
    !!              bal%blkweights_ps(0:nlocblocks), index 0 is unused, weights reside in 1:nlocblocks
    !!         - global number of blocks
    !!              nblocks
    !!         - number partitions (== number of ranks calling 1D partitioning == number of distributed weight vectors)
    !!              nproc
    !!         - current partition vector, contains start indexes of partitions/ranks in (global) weight vector
    !!              bal%partition(0:nproc), partition(0) = 1, partition(nproc) = nblocks + 1
    !! Output: - new partition vector
    !!              bal%partition(0:nproc), old is backed up in bal%partition_old(0:nproc)
    !!


    call fd4_balance_readjust(domain, err, opt_stats=bal_stats)
    
    if(err/=0) then
      write(out,*) 'fd4_balance_readjust failed'
      i = err
      call MPI_Abort(MPI_COMM_WORLD, i, err)
    end if
    

    partition => bal%partition
    partition_old => bal%partition_old
    !write(out,*) mpi%rank, domain%blocks%length, bal%localweight


    ! get max/min load, req. steps, max weight, min blk, maxweight
    if(method_ser(method) .or. method==4) then
      ! non-distributed methods: just get all the stats from rank 0
      if(mpi%rank==0) then
        maxload = 0.0_r8k
        min_blocks = nblocks
        do crank=0,mpi%nprocs-1
          maxload = MAX(maxload, bal%blkweights_ps(partition(crank+1)-1) - bal%blkweights_ps(partition(crank)-1) )
          tmp = partition(crank+1)-partition(crank)
          if(tmp<min_blocks) then
            min_blocks = tmp
            min_blocks_rank = crank
          end if
        end do
        sumload = bal%blkweights_ps(partition(mpi%nprocs)-1)
      end if
    else
      if(method==11 .or. method==12 .or. method==15 .or. method==16 .or. method==17 .or. method==18 .or. method==22) then
        ! bal%localweight not set
        ! only group master has blkweights_ps ... master needs to scatter localweight to all group members
        if(.not. allocated(localwgts)) allocate(localwgts(0:bal%sfcd_mpi%nprocs-1))
        if(bal%sfcd_mpi%rank==0) then
          do crank=0,bal%sfcd_mpi%nprocs-1
            wrank = crank + mpi%rank
            localwgts(crank) = &
              bal%blkweights_ps( partition(wrank+1)-partition(mpi%rank) ) &
            - bal%blkweights_ps( partition(wrank)  -partition(mpi%rank) )
          end do
        end if
        call MPI_Scatter(localwgts(0), 1, MPI_R8K, bal%localweight, 1, MPI_R8K, 0, bal%sfcd_mpi%comm, err)
      end if
      call MPI_Reduce(bal%localweight, sumload, 1, MPI_R8K, MPI_SUM, 0, MPI_COMM_WORLD, err)
      call MPI_Reduce(bal%localweight, maxload, 1, MPI_R8K, MPI_MAX, 0, MPI_COMM_WORLD, err)
#define DETAILED_STATS
#ifdef DETAILED_STATS
      call MPI_Allreduce(domain%blocks%length, min_blocks, 1, MPI_INTEGER, MPI_MIN, MPI_COMM_WORLD, err)
      tmp = mpi%nprocs
      if(domain%blocks%length==min_blocks) tmp=mpi%rank
      call MPI_Reduce(tmp, min_blocks_rank, 1, MPI_INTEGER, MPI_MIN, 0, MPI_COMM_WORLD, err)
#else
      min_blocks = -1
      min_blocks_rank = 0
#endif
    end if
#ifdef DETAILED_STATS
    tmp = INT(domain%stat%values(bal%stat_binsteps))
    call MPI_Reduce(tmp, req_steps, 1, MPI_INTEGER, MPI_MAX, 0, MPI_COMM_WORLD, err)    
    call MPI_Reduce(maxweight_loc, maxweight, 1, MPI_R8K, MPI_MAX, 0, MPI_COMM_WORLD, err)
    if(mpi%rank==0) then
      weight_balance = sumload / (maxweight * nblocks)
    end if
#else
    req_steps = INT(domain%stat%values(bal%stat_binsteps)) ! from rank 0
    maxweight = 0.0
    weight_balance = 0.0
#endif

    ! get timing (sum over all ranks)
    call MPI_Reduce(domain%stat%values, stat_values_coll, size(domain%stat%values), MPI_I8K, MPI_SUM, 0, MPI_COMM_WORLD, err)


    ! dump fancy stats!
    if(mpi%rank==0) then

      tsum = stat_values_coll(bal%stat_pfxs_us,1,1) &
           + stat_values_coll(bal%stat_exwgt_us,1,1) &
           + stat_values_coll(bal%stat_p1dc_us,1,1) &
           + stat_values_coll(bal%stat_bcbr_us,1,1) &
           + stat_values_coll(bal%stat_p1d_us,1,1) &
           + stat_values_coll(bal%stat_dspv_us,1,1)

      balance = sumload / (maxload * mpi%nprocs)

      if(dumpcounter==step_select(2)) write(out,'(A)') 'end of warm-up runs'
#if (defined(MEM_DUMP) && defined(__bgq__))
      ! BlueGene/Q
      call get_memory_heap(heap_usage)
#else
      heap_usage = 0.0_r8k
#endif
    
      write(out,'(I3,F10.7,1P,E10.3,I5,I7,0P,I3,A1,F4.3,I8,I11,F11.1,F11.1,F10.2,F10.2,F12.4,F13.1,I4)') &
               dumpcounter,balance, &
               maxload, & ! maxload
               min_blocks, min_blocks_rank, &  ! min.blk,  at
               req_steps, & ! max. req. steps
               ' ',weight_balance, &   ! wgt.bal
               bal_stats%migrated_blocks, &
               INT(REAL(tsum)/REAL(mpi%nprocs),i8k), &
               REAL(stat_values_coll(bal%stat_pfxs_us,1,1))/REAL(mpi%nprocs), &
               REAL(stat_values_coll(bal%stat_exwgt_us,1,1))/REAL(mpi%nprocs), &
               REAL(stat_values_coll(bal%stat_p1dc_us,1,1))/REAL(mpi%nprocs), &
               REAL(stat_values_coll(bal%stat_bcbr_us,1,1))/REAL(mpi%nprocs), &
               REAL(stat_values_coll(bal%stat_p1d_us,1,1))/REAL(mpi%nprocs), &
               REAL(stat_values_coll(bal%stat_dspv_us,1,1))/REAL(mpi%nprocs), &
               INT(heap_usage)

!#define FD4_SFC_DISTR_DEBUG_BALANCE
#ifdef FD4_SFC_DISTR_DEBUG_BALANCE
      write(9000+dumpcounter,'(I6,I12,I12,F16.4,F16.4,F14.5,F16.4,F14.5)') &
        dumpcounter, 1, nblocks, bal%blkweights_ps(nblocks), maxweight, &
        bal%blkweights_ps(nblocks)/(maxweight*nblocks), &
        maxload, balance
#endif

    end if
      

    if(dumpcounter>=step_select(2)) then

      ! do not use all cycles for summary stats (leave out first ones for warming up)
      ncycles = ncycles + 1

      ! get migration stats
      migrated_blocks = migrated_blocks + bal_stats%migrated_blocks
      partition_shifts = partition_shifts + sum(INT(abs(partition_old-partition),i8k))
      if(partition_shifts<migrated_blocks) then
        print *,sum(INT(abs(partition_old-partition),i8k))
        stop 'partition_shifts<migrated_blocks'
      end if

      if(mpi%rank==0) then
        avgbalance = avgbalance + balance
      end if

      ! save FD4 stats
      stat_values(:,dumpcounter) = domain%stat%values(:)
      
    end if

    
  end do
  
  deallocate(stat_values_coll)
  allocate(stat_values_coll(size(domain%stat%values),step_select(2):step_select(3),3))

  stat_values_coll = 0
  
  call MPI_Reduce(stat_values(1,step_select(2)), stat_values_coll(1,step_select(2),1), size(stat_values), &
                  MPI_I8K, MPI_MIN, 0, MPI_COMM_WORLD, err)
  call MPI_Reduce(stat_values(1,step_select(2)), stat_values_coll(1,step_select(2),2), size(stat_values), &
                  MPI_I8K, MPI_MAX, 0, MPI_COMM_WORLD, err)
  call MPI_Reduce(stat_values(1,step_select(2)), stat_values_coll(1,step_select(2),3), size(stat_values), &
                  MPI_I8K, MPI_SUM, 0, MPI_COMM_WORLD, err)

  if(mpi%rank==0) then
    write(out,'(A,I8)')    'nprocs     =',mpi%nprocs
    write(out,'(A,I8,A,I8,A)') 'nblocks    =',nblocks,' (orig:',PRODUCT(blnum_file),')'
    write(out,'(A,F9.2)')  'nblk/nprc  =',REAL(nblocks)/REAL(mpi%nprocs)
    write(out,'(A,I8)')    'io_groupsz =',io_group_size
    write(out,'(A,I8)')    'ncycles    =',ncycles
    write(out,'(A,2(I5))')    'steps      =',step_select(2:3)
    write(out,'(A,I8,A,A)')    'sfctype    =',sfctype,'   ',sfc_str(sfctype)
    write(out,'(A,I8,A,A)')    'method     =',method,'   ',method_str(method)
    write(out,'(A,I8)')        'groupsz    =',sfcd_groupsz
    write(out,'(A,I8)')        'ngroups    =',mpi%nprocs / sfcd_groupsz
    write(out,'(A,I8)')        'p1d_root   =',p1d_root
    write(out,'(A,F9.4)') 'lb_target  =',lb_target
    write(out,'(A,F9.4)') 'rndm_pert  =',rndm_perturbation
    write(out,'(A,F15.8)') 'avgbalance =',avgbalance/REAL(ncycles,r8k)
    write(out,'(A,F15.3)')   't pfxs     =',SUM(stat_values_coll(bal%stat_pfxs_us,:,3))/REAL(mpi%nprocs)
    write(out,'(A,F15.3)')   't xwgt     =',SUM(stat_values_coll(bal%stat_exwgt_us,:,3))/REAL(mpi%nprocs)
    write(out,'(A,F15.3)')   't p1dc     =',SUM(stat_values_coll(bal%stat_p1dc_us,:,3))/REAL(mpi%nprocs)
    write(out,'(A,F15.3)')   't bcbr     =',SUM(stat_values_coll(bal%stat_bcbr_us,:,3))/REAL(mpi%nprocs)
    write(out,'(A,F15.3)')   't p1d      =',SUM(stat_values_coll(bal%stat_p1d_us,:,3))/REAL(mpi%nprocs)
    write(out,'(A,F15.3)')   't dspv     =',SUM(stat_values_coll(bal%stat_dspv_us,:,3))/REAL(mpi%nprocs)
    write(out,'(A,F9.3)')  'avg_steps  =', SUM(stat_values_coll(bal%stat_binsteps,:,2))/REAL(ncycles)
    write(out,'(A,I13)')    'mig blocks =',migrated_blocks
    !write(out,'(A,I13)')    'mig blocks =',SUM(stat_values_coll(bal%stat_migblock,:,3))
    write(out,'(A,I13)')    'mig shifts =',partition_shifts
  end if
  
  !call MPI_Barrier(MPI_COMM_WORLD, err)
  !if(mpi%rank==0) then
  !  write(51,'(A8)') 'finished'
  !  close(51)
  !end if

  ! reconstruct FD4 stats
  domain%stat%values(:) = SUM(stat_values(:,step_select(2):step_select(3)), DIM=2)
  domain%stat%values(bal%stat_binsteps) = MAXVAL(stat_values(bal%stat_binsteps,step_select(2):step_select(3)))

  !call fd4_domain_delete(domain, opt_stats=.false.)
  call fd4_domain_delete(domain)
  if(io_comm /= MPI_COMM_WORLD .and. io_comm /= MPI_COMM_NULL) then
    call MPI_Comm_free(io_comm, err)
  end if
  call MPI_Finalize(err)

contains


  SUBROUTINE init_random_seed(seed_in)
         integer, intent(in) :: seed_in
         INTEGER :: n
         INTEGER, DIMENSION(:), ALLOCATABLE :: seed

         CALL RANDOM_SEED(size = n)
         ALLOCATE(seed(n))
         seed(:) = seed_in

         CALL RANDOM_SEED(PUT = seed)

         DEALLOCATE(seed)
  END SUBROUTINE


  function get_replicate_weight(blkweights3d, blnum, i, j, k) result(wgt)
    real(r8k), intent(in) :: blkweights3d(:,:,:)
    integer, intent(in) :: blnum(3), i, j, k
    real(r8k) :: wgt
    
    wgt = blkweights3d( MOD(i-1,blnum(1))+1, MOD(j-1,blnum(2))+1, MOD(k-1,blnum(3))+1 )
    
  end function get_replicate_weight
    

end program partbench
