! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                       2D Hilbert Curve                       !
!                                                              !
!--------------------------------------------------------------!

#ifdef SFC_ERRCHECK
#define PURE
#endif

#ifndef SFC_MAX_LEV
#define SFC_MAX_LEV 12
#endif

!! Simple 2D Hilbert curve subroutines.
!! Based on a table description of the curve.
!!
!! Routines accept position and hilbert-index starting at 1.
!!
!! @author Matthias Lieber
module hilbert2d_mod

  use kinds_mod
  implicit none
  private  
  
  public :: hilbert2d_i2c, hilbert2d_c2i
  public :: hilbert2d_i2c_qsz, hilbert2d_c2i_qsz
  public :: hilbert2d_iter, hilbert2d_iter_init, hilbert2d_iter_set, hilbert2d_iter_next
  
  !! Type for one entry in the 2D Hilbert curve description table.
  !!
  !! Values of the $mv item are:
  !! - 1/-1 means x+1/x-1 
  !! - 2/-2 means y+1/y-1 
  type hilbert_table
    integer(i_k) :: mv(0:2)      !! spatial moves of the curve
    integer(i_k) :: next(0:3)    !! next states when refinining (quadrant is sequence number)
    integer(i_k) :: num2ref(0:3) !! sequence number on curve primitve to reference (spatial) quadrant 
    integer(i_k) :: ref2num(0:3) !! reference (spatial) quadrant to sequence number on curve primitve
  end type hilbert_table  
  
  integer(i_k), parameter :: SFC_DIMS = 2      !! spatial dimensions of SFC
  
  type(hilbert_table), parameter :: tab(0:3) =                            &
  (/                                                                      &
    hilbert_table( (/ 2, 1,-2/), (/1,0,0,2/), (/0,2,3,1/), (/0,3,1,2/) ), &
    hilbert_table( (/ 1, 2,-1/), (/0,1,1,3/), (/0,1,3,2/), (/0,1,3,2/) ), &
    hilbert_table( (/-1,-2, 1/), (/3,2,2,0/), (/3,2,0,1/), (/2,3,1,0/) ), &
    hilbert_table( (/-2,-1, 2/), (/2,3,3,1/), (/3,1,0,2/), (/2,1,3,0/) )  &
   /)
  
  !! Iterator to traverse the 2D Hilbert curve
  !!
  type hilbert2d_iter
    integer(i_k) :: idx = - 1
    integer(i_k) :: p(SFC_DIMS) = - 1
    integer(i_k) :: lev = -1
    integer(i_k) :: qsz(SFC_DIMS) = -1
    integer(i_k) :: len = -1
    integer(i_k) :: clev = -1
    integer(i_k) :: s(SFC_MAX_LEV) = -1
    integer(i_k) :: c(SFC_MAX_LEV) = -1
    integer(i_K) :: jump_level = SFC_MAX_LEV + 1
  end type hilbert2d_iter
  
contains

include 'util/math.incf'

!! Get index on curve at given 2D coordinate (coordinates-to-index).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert2d_c2i(lev, p, idx)

  integer(i_k), intent(in)  :: lev    !! level of the hilbert curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: p(2)   !! 2D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 2D hilbert curve [1 ... (2^$lev)^2]
  
  integer(i_k) :: l,len,dim,q,s,rq,cp(2)
    
  cp(:)=p(:)-1
  idx=1
  s=0
  dim=2**lev
  ! length of curve in base quadrant
  len=dim**2

#ifdef SFC_ERRCHECK  
  ! very expensive check
  if(cp(1)>=dim .or. cp(1)<0 .or. cp(2)>=dim .or. cp(2)<0) then
    print *,'hilbert2d_c2i: position out of bounds', cp
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension and length of refined quadrants
    dim=dim/2
    len=len/4
    
    ! determine spatial quadrant and reduce spatial coordinates
    ! so that they are relative to refined quadrant
    ! reference quadrant:
    ! 2 3
    ! 0 1
    rq=0
    if(cp(1)>=dim) then
      rq=ior(1,rq)
      cp(1)=cp(1)-dim
    end if
    if(cp(2)>=dim) then
      rq=ior(2,rq)
      cp(2)=cp(2)-dim
    end if
    
    ! get sequence number of quadrant q in the curve primitive
    q=tab(s)%ref2num(rq)
    
    ! add offset to index    
    idx=idx+q*len
    
    ! get next state
    s=tab(s)%next(q)
    
  end do
  
end subroutine hilbert2d_c2i


!! Get 2D coordinate at given index on curve (index-to-coordinates).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert2d_i2c(lev, idx, p)

  integer(i_k), intent(in)  :: lev   !! level of the hilbert curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: idx   !! index on the 2D hilbert curve [1 ... (2^$lev)^2 ]
  integer(i_k), intent(out) :: p(2)  !! 2D coordinates [1 ... 2^$lev]

  integer(i_k) :: l,len,dim,q,cidx,s,rq
  
  p(:)=1
  s=0
  cidx=idx-1
  dim=2**lev
  
  ! length of curve in base quadrant
  len=dim**2

#ifdef SFC_ERRCHECK  
  if(cidx>=len .or. cidx<0) then
    print *,'hilbert2d_i2c: cidx >= len or cidx < 0: ', cidx
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension of refined quadrants
    dim=dim/2
    
    ! determine number of quadrant q in the curve primitive
    if(cidx<len/2) then
      if(cidx<len/4) then
        q=0
      else
        q=1
      end if
    else
      if(cidx<3*len/4) then
        q=2
      else
        q=3
      end if
    end if
    
    ! get actual spatial quadrant, depending on curve primitive
    ! reference quadrant:
    ! 2 3
    ! 0 1
    rq=tab(s)%num2ref(q)
    
    ! update coordinates
    if(iand(1,rq)==1) p(1)=p(1)+dim
    if(iand(2,rq)==2) p(2)=p(2)+dim
    
    ! length of curve in refined quadrant
    len=len/4
    
    ! decrease idx by offset in current quadrant
    cidx=cidx-q*len
    
    ! next state
    s=tab(s)%next(q)
    
  end do
 
end subroutine hilbert2d_i2c

!! Get index on curve at given 2D coordinate (coordinates-to-index).
!!
!! The size of the query region is restricted by $qsz for each dimension.
!! $qsz must be <= $lev**2 and $pos must be <= $qsz.
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert2d_c2i_qsz(lev, qsz, p, idx)

  integer(i_k), intent(in)  :: lev    !! level of the hilbert curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: qsz(2) !! query region size
  integer(i_k), intent(in)  :: p(2)   !! 2D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 2D hilbert curve [1 ... (2^$lev)^2]
  
  integer(i_k) :: l,dim,q,s,rq,cp(2)
  integer(i_k) :: csz(2), iq, irq, rsz(2)
    
  cp(:)=p(:)-1
  idx=1
  s=0
  dim=2**lev
  
  csz = qsz

#ifdef SFC_ERRCHECK  
  ! very expensive check
  if(cp(1)>=qsz(1) .or. cp(1)<0 .or. cp(2)>=qsz(2) .or. cp(2)<0) then
    print *,'hilbert2d_c2i: position out of bounds', cp
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension and length of refined quadrants
    dim=dim/2
    
    ! determine spatial quadrant and reduce spatial coordinates
    ! so that they are relative to refined quadrant
    ! reference quadrant:
    ! 2 3
    ! 0 1
    rq=0
    if(cp(1)>=dim) then
      rq=ior(1,rq)
      cp(1)=cp(1)-dim
    end if
    if(cp(2)>=dim) then
      rq=ior(2,rq)
      cp(2)=cp(2)-dim
    end if
    
    ! get sequence number of quadrant q in the curve primitive
    q=tab(s)%ref2num(rq)

    ! add offset to index
    do iq=0,q-1
      ! count number of cells per quadrant that come before the current quadrant q
      irq = tab(s)%num2ref(iq)
      if(iand(irq,1)==0) then
        rsz(1) = min(dim, csz(1))
      else
        rsz(1) = max(0  , csz(1)-dim)
      end if
      if(iand(irq,2)==0) then
        rsz(2) = min(dim, csz(2))
      else
        rsz(2) = max(0  , csz(2)-dim)
      end if
      idx = idx + PRODUCT(rsz)
    end do
    
    ! shrink the current query size (csz) relative to current quadrant
    if(iand(rq,1)>0) then
      csz(1) = csz(1) - dim
    else
      csz(1) = min(csz(1), dim)
    end if
    if(iand(rq,2)>0) then
      csz(2) = csz(2) - dim
    else
      csz(2) = min(csz(2), dim)
    end if
    
    ! get next state
    s=tab(s)%next(q)
    
  end do
  
end subroutine hilbert2d_c2i_qsz


!! Get 2D coordinate at given index on curve (index-to-coordinates).
!!
!! The size of the query region is restricted by $qsz for each dimension.
!! $qsz must be <= $lev**2 and $pos must be <= $qsz.
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine hilbert2d_i2c_qsz(lev, qsz, idx, p, iter)

  integer(i_k), intent(in)  :: lev   !! level of the hilbert curve (1=2x2, 2=4x4, 3=8x8, etc)
  integer(i_k), intent(in)  :: qsz(2)!! query region size
  integer(i_k), intent(in)  :: idx   !! index on the 2D hilbert curve [1 ... (2^$lev)^2 ]
  integer(i_k), intent(out) :: p(2)  !! 2D coordinates [1 ... 2^$lev]
  type(hilbert2d_iter), optional, intent(inout) :: iter !! for internal use only

  integer(i_k) :: l,dim,q,cidx,s,rq
  integer(i_k) :: csz(2), iq, irq, rsz(2), newcidx
  
  p(:)=1
  s=0
  cidx=idx-1
  dim=2**lev

  csz = qsz

#ifdef SFC_ERRCHECK  
  if(cidx>=PRODUCT(qsz) .or. cidx<0) then
    print *,'hilbert2d_i2c: cidx >= PRODUCT(qsz) or cidx < 0: ', cidx
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension of refined quadrants
    dim=dim/2
    
    ! decrease cidx by offset in current quadrant
    ! and determine sequence number q of quadrant in the curve primitive.
    q=3
    do iq=0,2
      ! count number of cells per quadrant that come before the current quadrant q
      irq=tab(s)%num2ref(iq)
      if(iand(irq,1)==0) then
        rsz(1) = min(dim, csz(1))
      else
        rsz(1) = max(0  , csz(1)-dim)
      end if
      if(iand(irq,2)==0) then
        rsz(2) = min(dim, csz(2))
      else
        rsz(2) = max(0  , csz(2)-dim)
      end if
      newcidx = cidx - PRODUCT(rsz)
      if(newcidx>=0) then
        cidx = newcidx
      else
        q=iq
        exit
      end if
    end do

    ! get actual spatial quadrant, depending on curve primitive
    ! reference quadrant:
    ! 2 3
    ! 0 1
    rq=tab(s)%num2ref(q)
    
    ! shrink the current query size (csz) relative to current quadrant
    ! and update coordinates
    if(iand(rq,1)>0) then
      p(1)=p(1)+dim
      csz(1) = csz(1) - dim
    else
      csz(1) = min(csz(1), dim)
    end if
    if(iand(rq,2)>0) then
      p(2)=p(2)+dim
      csz(2) = csz(2) - dim
    else
      csz(2) = min(csz(2), dim)
    end if

    if(present(iter)) then
      iter%s(l)=s
      iter%c(l)=q
    end if

    ! next state
    s=tab(s)%next(q)
    
  end do

  if(present(iter)) then
    iter%clev = 1
    do while(iter%c(iter%clev)>2)
      ! go up one level
      iter%clev = iter%clev + 1
      !print *,' next level+',iter%clev
    end do
  end if 
  
end subroutine hilbert2d_i2c_qsz



!! Initialize the iterator.
!!
!! The current index is $iter%$idx and the current 2D position $iter%$p.
!!
!!> call hilbert2d_iter_init(iter, lev, qsz)
!!  do while(iter%idx > 0)
!!    array(iter%p(1),iter%p(2)) = iter%idx
!!    call hilbert2d_iter_next(iter)
!!  end do
!!< 
!!
PURE subroutine hilbert2d_iter_init(iter, lev, opt_qsz)

  type(hilbert2d_iter), intent(inout) :: iter
  integer(i_k), intent(in) :: lev
  integer(i_k), intent(in), optional :: opt_qsz(SFC_DIMS)
  
  integer :: l
  
  if(lev>SFC_MAX_LEV) then
    iter%idx = -1
#ifdef SFC_ERRCHECK  
    print *,'hilbert2d_iter_init: lev > SFC_MAX_LEV = ',SFC_MAX_LEV
#endif
    return
  end if
  
  iter%idx = 1
  iter%p = 1
  iter%c = 0
  iter%lev = lev
  if(present(opt_qsz)) then
    iter%qsz = opt_qsz
    ! define the level at which we start jumping depending on the smallest dimension of the query size
    iter%jump_level = MAX( log2i(MINVAL(iter%qsz)) - 2, 1)
  else
    iter%qsz = 2**lev
    iter%jump_level = SFC_MAX_LEV + 1
  end if
  iter%len = PRODUCT(iter%qsz)
  
  iter%clev = 1
  iter%s(iter%lev) = 0
  do l=iter%lev,2,-1
    iter%s(l-1) = tab(iter%s(l))%next(0)
  end do
  
end subroutine hilbert2d_iter_init


!! Set the iterator to given index
!!
PURE subroutine hilbert2d_iter_set(iter, idx)
  type(hilbert2d_iter), intent(inout) :: iter
  integer(i_k), intent(in) :: idx
  
  iter%idx = idx
  call hilbert2d_i2c_qsz(iter%lev, iter%qsz, iter%idx, iter%p, iter)
  
end subroutine hilbert2d_iter_set


!! Iterate to next point on the hilbert curve
!!
!! After reaching the end of the curve, $iter%$idx is set to 0.
!!
PURE subroutine hilbert2d_iter_next(iter)

  type(hilbert2d_iter), intent(inout) :: iter
  
  integer(i_k) :: s, c, l
  logical :: outside_qsz
  
  iter%idx = iter%idx + 1
  
  do

    if(iter%idx>iter%len .or. iter%clev>iter%lev) then
      ! end of iteration
      iter%idx = 0
      exit
    end if

    s=iter%s(iter%clev)
    c=iter%c(iter%clev)
    iter%p(abs(tab(s)%mv(c))) = iter%p(abs(tab(s)%mv(c))) + sign(1,tab(s)%mv(c)) 
    !print *,iter%p,' |',iter%idx,iter%clev,c,s
    
    outside_qsz = ANY(iter%p>iter%qsz)

    if(outside_qsz .and. iter%clev>iter%jump_level) then
    !if(.false.) then

      ! jump if new point out of query region
      call hilbert2d_i2c_qsz(iter%lev, iter%qsz, iter%idx, iter%p, iter)
      !print *,' jumping to ', iter%idx,' at',iter%p
      !print *,'      *c*',iter%c(1:iter%lev)
      !print *,'      *s*',iter%s(1:iter%lev)
      outside_qsz = .false.

    else ! not jumping: iterate to next point

      iter%c(iter%clev) = iter%c(iter%clev) + 1
      if(iter%clev>1) then
        do while(iter%clev>1)
          ! go down one level
          l = iter%clev - 1
          !print *,' next level-',l,iter%s(iter%clev),iter%c(iter%clev)
          iter%s(l) = tab(iter%s(iter%clev))%next(iter%c(iter%clev))
          iter%c(l) = 0
          iter%clev = l
        end do
      else
        !iter%c(iter%clev) = iter%c(iter%clev) + 1
        do while(iter%c(iter%clev)>2)
          ! go up one level
          iter%clev = iter%clev + 1
          !print *,' next level+',iter%clev
        end do
      end if

    end if ! not jumping
    
    if(.not.outside_qsz) exit
    !iter%idx = iter%idx - 1 ; exit
    
  end do
  

end subroutine hilbert2d_iter_next


end module hilbert2d_mod
